--[[
--
--  EkBomb 5.01
--
--  Refer to index.html for documentation and complete change history.
--
--  Author: Dargen of Eternal Keggers, Terenas.
--          http://www.eternalkeggers.net
--
--  Addons: http://www.wowinterface.com/portal.php?uid=25131
--          http://my.curse.com/dargen/projects/
--
--
--  Calling EkBomb functions from another addon:
--
--	To show a custom bomb warning from another addon, you can
--	call this function:
--
--		EkBomb_Show("bigText", "smallText", seconds)
--
--	All three parameters are optional. Default values will be used for missing
--	or nil parameters.  Both of the default messages are blank, and the default
--	number of seconds is 30. To prevent the warning window from automatically
--	closing, use 0 seconds.
--
--	If you want to hide the bomb warning, you can call this function:
--
--		EkBomb_Hide()
--
--]]

--!!
--&&
--^^

-- Saved variables
EkBomb_Status = 1;   -- 0==Warning window disabled, 1==Warning window enabled.
EkBomb_Opacity = 80;  -- Warning window text opacity
EkBomb_BombsConfig = {};  -- User settings for the predefined warnings table.
EkBomb_FlashWindow = 0;  -- 0=Window does not flash, 1==Window flashes.
EkBomb_Sound = 1;  -- 0==Don't play a sound, 1==Play a sound when warning appears.

-- New saved variables in v1.3
EkBomb_Background = 0;  -- 0==Don't show background, 1==Show background
EkBomb_Scale = 100;  -- Warning window scale %
EkBomb_Color = nil;  -- colors

-- New saved variables in v1.41
EkBomb_ShowTimer = 0;  -- 0==Don't show timer on close button, 1==Show timer

-- New saved variables in v2.00
EkBomb_WarningPosition = nil;  -- Saved position and scale of the warning frame
EkBomb_EventPosition = nil;  -- Saved position of the event frame
EkBomb_CustomTable1 = {};  -- List of custom warnings (general -- applies to all characters).
EkBomb_CustomTable2 = {};  -- List of custom warnings (character specific).
EkBomb_EventBufferSize = 500;  -- Size of the circular buffer (number of events)
EkBomb_ShowLink = 0;  -- 1==Record chat events and show event link on chat line, 0==Don't.
EkBomb_ShowTooltips = 1;  -- 1==Show detailed tooltips, 0=Don't.
EkBomb_ShowAtCursor = 0;  -- 1==Show detailed tooltips by the mouse pointer, 0==Show at default location (lower right corner of screen).


-- Variables that are not saved
EkBomb_Version = "5.01";

EkBomb_Slash = "/ekbomb";
EkBomb_Addon = "EkBomb";

EkBomb_WarningTimer = nil;
EkBomb_WarningTimer2 = 0.1;
EkBomb_FlashTimer = nil;
EkBomb_FlashState = 1;
EkBomb_TestFlag = 0;
EkBomb_PauseWarningTimer = nil;  -- 1==Pause warning frame timer, nil==Count down

EkBomb_CloseButton_Format = EkBomb_TEXT_Warning_Close .. "  %d";

EkBomb_TEXT_Config_Help = EkBomb_TEXT_Config_Help .. " " .. EkBomb_Slash .. " " .. EkBomb_TEXT_Command_Help;

EkBomb_TestMode = nil;  -- 1==User pressed test button to show current warning, nil==User did not press test button.
EkBomb_ScaleOverride = nil;  -- 1==Warning window did not fit so we are overriding the scale, nil==Warning window fit on screen.
EkBomb_ScaleOnShow = nil;  -- Set flag to tell EkBomb_OptionsFrame_Scale_OnValueChanged() that this routine changed the slider's value.

EkBomb_TestList  = 1;  -- Used when testing warnings: List number to be used
EkBomb_TestIndex = 0;  -- Used when testing warnings: Subscript in list to use

EkBomb_ChatFrame_Old = {};  -- Used to save chat frames' AddMessage value when hooking them.
EkBomb_SetItemRef_Old = nil;  -- Used to save SetItemRef() value when hooking it.

-- Frames that are displayed within the EkBomb_MainFrame frame when a tab is clicked on.
EkBomb_MainFrame_SUBFRAMES = {
	"EkBomb_OptionsFrame",
	"EkBomb_PredefinedFrame",
--!!	"EkBomb_CustomFrame",
};

-- Main tabs
EkBomb_MainFrame_TabOptions = 1;
EkBomb_MainFrame_TabPredefined = 2;
EkBomb_MainFrame_TabCustom = 3;

-- Tabs on the list frames (pre-defined frame only uses the general tab)
EkBomb_ListFrame_TabGeneral = 1;
EkBomb_ListFrame_TabCharacter = 2;

EkBomb_ListFrame_Table = nil;  -- Reference to current list (being viewed).
EkBomb_ListFrame_TableNum = nil;  -- Table number (1==CustomTable2, 2==CustomTable1, 3==PredefinedTable1. See EkBomb_WarningLists variable)
EkBomb_ListFrame_NumToDisplay = nil;  -- Number of items to display vertically
EkBomb_ListFrame_Selected = 0;  -- Selected item subscript (0==Nothing selected).
EkBomb_ListFrame_MenuNum = 1; -- Current menu number
EkBomb_ListFrame_SavedScrollValue = nil;
EkBomb_ListFrame_Tooltip_OffsetRight = 25;  -- Offset when displaying tooltip on the right (varies).

-- EkBomb_ListFrame_Saved[] -- Holds some values from last time a list was displayed.
--   - Subscript 1 is the name of the list ("PRE1", "CUS1", "CUS2").
--   - Subscript 2 is a code for the value being saved.
--   - Refer to: EkBomb_ListFrame_SaveValues() and EkBomb_ListFrame_RestoreValues()
EkBomb_ListFrame_Saved = {};

EkBomb_EditFrame_EB_MaxWidth = nil;  -- Max width of editboxes on EkBomb_EditFrame 
EkBomb_EditFrame_EditMode = 0;  -- 0==View, 1==Edit, 2==Add
EkBomb_EditFrame_Clipboard = nil;  -- Used to hold warning properties for use with copy/paste buttons

-- List of editboxes used to determine Tab key sequence.
EkBomb_EditFrame_EB_List = {
	"EkBomb_EditFrame_NameEB",
	"EkBomb_EditFrame_SmallEB",
	"EkBomb_EditFrame_LargeEB",
	"EkBomb_EditFrame_TimerEB",
	"EkBomb_EditFrame_EventEB",
	"EkBomb_EditFrame_ArgEB",
	"EkBomb_EditFrame_SearchValueEB",
	"EkBomb_EditFrame_ZoneEB",
};

-- Search codes
EkBomb_SearchCode_Exact = 1;
EkBomb_SearchCode_Pattern = 2;
EkBomb_SearchCode_Filter = 3;

EkBomb_EditFrame_SearchType_Table = {
	{ name = EkBomb_TEXT_SearchType_Exact, code = EkBomb_SearchCode_Exact },
	{ name = EkBomb_TEXT_SearchType_Pattern, code = EkBomb_SearchCode_Pattern },
};

-- Filter codes (for the COMBAT_LOG_EVENT_UNFILTERED flag fields)
EkBomb_FilterCode_Me = 1;
EkBomb_FilterCode_Mine = 2;
EkBomb_FilterCode_MyPet = 3;
EkBomb_FilterCode_FriendlyUnits = 4;
EkBomb_FilterCode_HostilePlayers = 5;
EkBomb_FilterCode_HostileUnits = 6;
EkBomb_FilterCode_NeutralUnits = 7;
EkBomb_FilterCode_UnknownUnits = 8;
EkBomb_FilterCode_Everything = 9;

EkBomb_FilterBits = {
	[EkBomb_FilterCode_Me] = COMBATLOG_FILTER_ME,
	[EkBomb_FilterCode_Mine] = COMBATLOG_FILTER_MINE,
	[EkBomb_FilterCode_MyPet] = COMBATLOG_FILTER_MY_PET,
	[EkBomb_FilterCode_FriendlyUnits] = COMBATLOG_FILTER_FRIENDLY_UNITS,
	[EkBomb_FilterCode_HostilePlayers] = COMBATLOG_FILTER_HOSTILE_PLAYERS,
	[EkBomb_FilterCode_HostileUnits] = COMBATLOG_FILTER_HOSTILE_UNITS,
	[EkBomb_FilterCode_NeutralUnits] = COMBATLOG_FILTER_NEUTRAL_UNITS,
	[EkBomb_FilterCode_UnknownUnits] = COMBATLOG_FILTER_UNKNOWN_UNITS,
	[EkBomb_FilterCode_Everything] = COMBATLOG_FILTER_EVERYTHING,
};


-- Classes (the code values must not be localized)
EkBomb_EditFrame_Class_Table = {
	{ name = EkBomb_TEXT_Class_Any,           code = "ANY" },
	{ name = EkBomb_TEXT_Class_DeathKnight,   code = "DEATHKNIGHT" },
	{ name = EkBomb_TEXT_Class_Druid,         code = "DRUID" },
	{ name = EkBomb_TEXT_Class_Hunter,        code = "HUNTER" },
	{ name = EkBomb_TEXT_Class_Mage,          code = "MAGE" },
	{ name = EkBomb_TEXT_Class_Monk,          code = "MONK" },
	{ name = EkBomb_TEXT_Class_Paladin,       code = "PALADIN" },
	{ name = EkBomb_TEXT_Class_Priest,        code = "PRIEST" },
	{ name = EkBomb_TEXT_Class_Rogue,         code = "ROGUE" },
	{ name = EkBomb_TEXT_Class_Shaman,        code = "SHAMAN" },
	{ name = EkBomb_TEXT_Class_Warlock,       code = "WARLOCK" },
	{ name = EkBomb_TEXT_Class_Warrior,       code = "WARRIOR" },
};

EkBomb_EventFrame_Buffer = {};  -- Circular buffer used to hold chat frame events
EkBomb_EventFrame_LastEventNum = 0;  -- Last event number in the buffer (starts at 0 each reload of the UI and keeps increasing by 1 per event)
EkBomb_EventFrame_FirstEventNum = 0; -- First event number in the buffer

-- Default colours for the small warning text, large warning text, and warning background.
EkBomb_DefaultColor = {
	["small"] = {
		["r"] = 0.0,
		["g"] = 0.22,
		["b"] = 1.0,
	},
	["large"] = {
		["r"] = 1.0,
		["g"] = 0.03137,
		["b"] = 0.08235,
	},
	["back"] = {
		["r"] = 0.6,
		["g"] = 0.6,
		["b"] = 0.6,
	},
};


-- List of pre-defined warnings
-- (The "id" keys are unique to this table. Don't add to the custom tables.)
-- (The "id" is used to look up user settings stored in EkBomb_BombsConfig)
--
-- COMBAT_LOG_EVENT_UNFILTERED arguments:
--
-- 4.1 4.2 <-- WoW version
--
--   1   1 == timestamp
--   2   2 == event
--   3   3 == hideCaster (new value Blizzard inserted in WoW 4.1)
--   4   4 == sourceGUID
--   5   5 == sourceName
--   6   6 == sourceFlags
--       7 == sourceRaidFlags (new value Blizzard inserted in WoW 4.2)
--   7   8 == destGUID
--   8   9 == destName
--   9  10 == destFlags
--      11 == destRaidFlags (new value Blizzard inserted in WoW 4.2)
--
-- For events that start with "SPELL":
--
--  10  12 == spellID
--  11  13 == spellName
--  12  14 == spellSchool

local COMBATLOG_ARG_event = 2;
local COMBATLOG_ARG_sourceName = 5;
local COMBATLOG_ARG_destFlags = 10;
local COMBATLOG_ARG_spellName = 13;

EkBomb_PredefinedTable1 = {
	{
		-- Anubisath Defender (Temple of Ahn'Qiraj):
		enabled = 1,
		id      = "anubisath",
		name    = EkBomb_TEXT_Bomb_Plague_Name,
		small   = EkBomb_TEXT_Bomb_Plague_Small,
		large   = EkBomb_TEXT_Bomb_Plague_Big,
		zone    = EkBomb_TEXT_Bomb_Plague_Zone,
		timer   = 30,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_AURA_APPLIED",
			},
			{
				arg    = COMBATLOG_ARG_destFlags,
				search = EkBomb_SearchCode_Filter,
				value  = EkBomb_FilterCode_Me,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Plague_Text,
			},
		},
	},
	{
		-- Shade of Aran (Karazhan):
		enabled = 1,
		id      = "aran",
		name    = EkBomb_TEXT_Bomb_Aran_Name,
		small   = EkBomb_TEXT_Bomb_Aran_Small,
		large   = EkBomb_TEXT_Bomb_Aran_Big,
		zone    = EkBomb_TEXT_Bomb_Aran_Zone,
		timer   = 26,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_CAST_START",
			},
			{
				arg    = COMBATLOG_ARG_sourceName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Aran_Mob,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Aran_Text,
			},
		},
	},
	{
		-- Baron Geddon (Molten Core):
		enabled = 1,
		id      = "geddon",
		name    = EkBomb_TEXT_Bomb_Living_Name,
		small   = EkBomb_TEXT_Bomb_Living_Small,
		large   = EkBomb_TEXT_Bomb_Living_Big,
		zone    = EkBomb_TEXT_Bomb_Living_Zone,
		timer   = 8,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_AURA_APPLIED",
			},
			{
				arg    = COMBATLOG_ARG_destFlags,
				search = EkBomb_SearchCode_Filter,
				value  = EkBomb_FilterCode_Me,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Living_Text,
			},
		},
	},
	{
		-- Grobbulus (Naxxramas):
		enabled = 1,
		id      = "grobbulus",
		name    = EkBomb_TEXT_Bomb_Mutating_Name,
		small   = EkBomb_TEXT_Bomb_Mutating_Small,
		large   = EkBomb_TEXT_Bomb_Mutating_Big,
		zone    = EkBomb_TEXT_Bomb_Mutating_Zone,
		timer   = 10,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_AURA_APPLIED",
			},
			{
				arg    = COMBATLOG_ARG_destFlags,
				search = EkBomb_SearchCode_Filter,
				value  = EkBomb_FilterCode_Me,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Mutating_Text,
			},
		},
	},
	{
		-- Bloodlord Mandokir (Zul'Gurub):
		enabled = 1,
		id      = "mandokir",
		name    = EkBomb_TEXT_Bomb_Mandokir_Name,
		small   = EkBomb_TEXT_Bomb_Mandokir_Small,
		large   = EkBomb_TEXT_Bomb_Mandokir_Big,
		zone    = EkBomb_TEXT_Bomb_Mandokir_Zone,
		timer   = 10,
		class   = "ANY",  -- Do not localize
		event   = "CHAT_MSG_MONSTER_YELL",
		args    = {
			{
				arg     = 1,
				search  = EkBomb_SearchCode_Pattern,
				value   = EkBomb_TEXT_Bomb_Mandokir_Player,
			},
		},
	},
	{
		-- Lurker Below (Serpentshrine Cavern):
		enabled = 1,
		id      = "lurker",
		name    = EkBomb_TEXT_Bomb_Lurker_Name,
		small   = EkBomb_TEXT_Bomb_Lurker_Small,
		large   = EkBomb_TEXT_Bomb_Lurker_Big,
		zone    = EkBomb_TEXT_Bomb_Lurker_Zone,
		timer   = 23,
		class   = "ANY",  -- Do not localize
		event   = "CHAT_MSG_RAID_BOSS_EMOTE",
		args    = {
			{
				arg     = 1,
				search  = EkBomb_SearchCode_Exact,
				value   = EkBomb_TEXT_Bomb_Lurker_Text,
			},
		},
	},
	{
		-- Nefarian (Blackwing Lair):
		enabled = 1,
		id      = "nefarian",
		name    = EkBomb_TEXT_Bomb_Corrupt_Name,
		small   = EkBomb_TEXT_Bomb_Corrupt_Small,
		large   = EkBomb_TEXT_Bomb_Corrupt_Big,
		zone    = EkBomb_TEXT_Bomb_Corrupt_Zone,
		timer   = 30,
		class   = "PRIEST",  -- Do not localize
		event   = "CHAT_MSG_MONSTER_YELL",
		args    = {
			{
				arg     = 1,
				search  = EkBomb_SearchCode_Exact,
				value   = EkBomb_TEXT_Bomb_Corrupt_Text,
			}
		},
	},
	{
		-- Vaelastrasz (Blackwing Lair):
		enabled = 1,
		id      = "vael",
		name    = EkBomb_TEXT_Bomb_Adrenaline_Name,
		small   = EkBomb_TEXT_Bomb_Adrenaline_Small,
		large   = EkBomb_TEXT_Bomb_Adrenaline_Big,
		zone    = EkBomb_TEXT_Bomb_Adrenaline_Zone,
		timer   = 20,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_AURA_APPLIED",
			},
			{
				arg    = COMBATLOG_ARG_destFlags,
				search = EkBomb_SearchCode_Filter,
				value  = EkBomb_FilterCode_Me,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Adrenaline_Text,
			},
		},
	},
	{
		-- Solarian (Tempest Keep: The Eye):
		enabled = 1,
		id      = "solarian",
		name    = EkBomb_TEXT_Bomb_Solarian_Name,
		small   = EkBomb_TEXT_Bomb_Solarian_Small,
		large   = EkBomb_TEXT_Bomb_Solarian_Big,
		zone    = EkBomb_TEXT_Bomb_Solarian_Zone,
		timer   = 6,
		class   = "ANY",  -- Do not localize
		event   = "COMBAT_LOG_EVENT_UNFILTERED",
		args    = {
			{
				arg    = COMBATLOG_ARG_event,
				search = EkBomb_SearchCode_Exact,
				value  = "SPELL_AURA_APPLIED",
			},
			{
				arg    = COMBATLOG_ARG_destFlags,
				search = EkBomb_SearchCode_Filter,
				value  = EkBomb_FilterCode_Me,
			},
			{
				arg    = COMBATLOG_ARG_spellName,
				search = EkBomb_SearchCode_Exact,
				value  = EkBomb_TEXT_Bomb_Solarian_Text,
			},
		},
	},
	{
		-- Sartharion (Obsidian Sanctum):
		enabled = 1,
		id      = "sartharion",
		name    = EkBomb_TEXT_Bomb_Sartharion_Name,
		small   = EkBomb_TEXT_Bomb_Sartharion_Small,
		large   = EkBomb_TEXT_Bomb_Sartharion_Big,
		zone    = EkBomb_TEXT_Bomb_Sartharion_Zone,
		timer   = 2,
		class   = "ANY",  -- Do not localize
		event   = "CHAT_MSG_RAID_BOSS_EMOTE",
		args    = {
			{
				arg     = 1,
				search  = EkBomb_SearchCode_Exact,
				value   = EkBomb_TEXT_Bomb_Sartharion_Text,
			},
			{
				arg     = 2,
				search  = EkBomb_SearchCode_Exact,
				value   = EkBomb_TEXT_Bomb_Sartharion_Mob,
			},
		},
	},
};

-- EkBomb_PredefinedTable2 = {};  -- Not being used.

-- EkBomb_EventList is a table where the key is an event name and the value is
-- another table. This second table contains a ["count"] element (number of warnings using this
-- event), and one [list_number] element for each list that contains a warning using this event.
-- Each [list_number] element is a table that contains the subscript numbers of the warnings
-- that use this event. The lists that the subscripts numbers refer to are the lists found
-- in the EkBomb_WarningLists table. The term list_number refers to the list's position in
-- the EkBomb_WarningLists table.
--
-- For example:
--
-- EkBomb_EventList={
--    ["CHAT_MSG_SPELL_PERIODIC_SELF_DAMAGE"]={  -- Event name
--       ["count"]=3,   -- Number of warnings using this event name
--       [2]={          -- List number 2 (ie. EkBomb_CustomTable1)
--          [1]=2,      -- The first warning in this list that uses this event name has a subscript value of 2 (ie. EkBomb_CustomTable1[2])
--       },
--       [3]={          -- List number 3 (ie. EkBomb_PredefinedTable1)
--          [1]=1,      -- The first  warning in this list that uses this event name has a subscript value of 1 (ie. EkBomb_PredefinedTable1[1])
--          [2]=3,      -- The second warning in this list that uses this event name has a subscript value of 3 (ie. EkBomb_PredefinedTable1[3])
--       },
--    },
--    ["CHAT_MSG_MONSTER_YELL"]={  -- Event name
--       ["count"]=2,   -- Number of warnings using this event name
--       [3]={          -- List number 3 (ie. EkBomb_PredefinedTable1)
--          [1]=2,      -- The first  warning in this list that uses this event name has a subscript value of 2 (ie. EkBomb_PredefinedTable1[2])
--          [2]=4,      -- The second warning in this list that uses this event name has a subscript value of 4 (ie. EkBomb_PredefinedTable1[4])
--       },
--    },
-- }
--
EkBomb_EventList = {};


EkBomb_EkCheck_Config = {
	addonName = "EkBomb",
	funcPrefix = "EkBomb_EkCheck_",
	defaultColumn = 11,
	columns = {
		{ heading=EkBomb_TEXT_EkCheck_Version,  width=56, show=11, sort=11, order="A", justify="CENTER", red=nil, green=nil, blue=nil },
	},
};

-- -----------
-- Blizzard addon code called from this addon:
--
--   GameTooltip.lua:
--     - GameTooltip_AddNewbieTip
--
--   UIDropDownMenu.lua:
--     - UIDropDownMenu_AddButton
--     - UIDropDownMenu_GetSelectedID
--     - UIDropDownMenu_Initialize
--     - UIDropDownMenu_SetSelectedID
--     - UIDropDownMenu_SetText
--     - UIDropDownMenu_SetWidth
--
--   UIPanelTemplates.lua:
--     - PanelTemplates_SetNumTabs
--     - PanelTemplates_SetTab
--     - PanelTemplates_Tab_OnClick
--     - PanelTemplates_TabResize
--     - PanelTemplates_UpdateTabs
--
--
-- Blizzard addon code which calls this addon:
--
--   UIDropDownMenu.lua:
--     - UIDropDownMenuButton_OnClick
--
--
-- Blizzard addon code hooked by this addon:
--
--   ChatFrame.lua
--     - ChatFrame?.AddMessage() (where ? == 1 to NUM_CHAT_WINDOWS)
--
--   ItemRef.lua:
--     - SetItemRef
--
--
-- Misc Blizzard functions called by this addon:
--
--     - CombatLog_Object_IsA
--
--
-- Blizzard variables used by this addon:
--
--     - COMBATLOG_FILTER_ME
--     - COMBATLOG_FILTER_MINE
--     - COMBATLOG_FILTER_MY_PET
--     - COMBATLOG_FILTER_FRIENDLY_UNITS
--     - COMBATLOG_FILTER_HOSTILE_PLAYERS
--     - COMBATLOG_FILTER_HOSTILE_UNITS
--     - COMBATLOG_FILTER_NEUTRAL_UNITS
--     - COMBATLOG_FILTER_UNKNOWN_UNITS
--     - COMBATLOG_FILTER_EVERYTHING
--     - DEFAULT_CHAT_FRAME
--     - HIGHLIGHT_FONT_COLOR
--     - NORMAL_FONT_COLOR
--     - NUM_CHAT_WINDOWS
--     - SlashCmdList
--     - UISpecialFrames
--
--
-- Blizzard frames used by this addon:
--
--     - UIParent
--     - ChatFrame? (where ? == 1 to NUM_CHAT_WINDOWS)
--
--
-- EkCheck references:
--
--     - EkCheck_RegisterAddon
-- -----------

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* Command line
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_ShowVersion(msg)
	-- --------------
	-- Show addon version information.
	-- --------------
	EkBomb_Print2(EkBomb_Addon .. " " .. EkBomb_Version .. " by Dargen, Eternal Keggers, Terenas.");
end

function EkBomb_Main(msg)
	-- --------------
	-- Main command routine.
	-- --------------
	local command = strlower(msg);

	if (command == "?" or command == EkBomb_TEXT_Command_Help) then
		-- Display help.
	        EkBomb_Print2(" ");
		EkBomb_ShowVersion()
	        EkBomb_Print2(" ");
	        EkBomb_Print2(EkBomb_Slash .. " -- " .. EkBomb_TEXT_Help_Config);
	        EkBomb_Print2(EkBomb_Slash .. " " .. EkBomb_TEXT_Command_Help .. " -- " .. EkBomb_TEXT_Help_Help);
	        EkBomb_Print2(EkBomb_Slash .. " " .. EkBomb_TEXT_Command_Reset .. " -- " .. EkBomb_TEXT_Help_Reset);
	        EkBomb_Print2(EkBomb_Slash .. " " .. EkBomb_TEXT_Command_Test .. " -- " .. EkBomb_TEXT_Help_Test);

	elseif (command == nil or command == "") then
		if (not EkBomb_MainFrame:IsShown()) then
			EkBomb_MainFrame:Show();
		else
			EkBomb_MainFrame:Hide();
		end

	elseif (command == EkBomb_TEXT_Command_Reset) then
		EkBomb_ResetWindows();

	elseif (command == EkBomb_TEXT_Command_Test) then
		EkBomb_TestWarning(1, nil, nil, nil, 0);

	elseif (command == "_test") then
		EkBomb_TestCommand();

	end
end

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_OnLoad(self)
	-- ----------------
	-- Addon loaded.
	-- ----------------
	SLASH_ekbomb1 = EkBomb_Slash;
	SlashCmdList["ekbomb"] = function( msg )
		EkBomb_Main(msg);
	end

	self:RegisterEvent("PLAYER_LOGIN");
end

function EkBomb_OnEvent(self, event, ...)
	-- ----------------
	-- Event routine
	-- ----------------
	if (event == "PLAYER_LOGIN") then

		-- EkBomb_WarningLists is an array holding references to each of the warning lists.
		-- The lists are in the sequence I want them to be searched when trying to match
		-- an event (custom per character, custom general, pre-defined general).
		--
		-- We need to initialize this table here after saved variables have been read in
		-- since the two custom tables are saved variables.
		EkBomb_WarningLists = { EkBomb_PredefinedTable1 };  --&&D
--!!		EkBomb_WarningLists = { EkBomb_CustomTable2, EkBomb_CustomTable1, EkBomb_PredefinedTable1 };

		if (not EkBomb_Color) then
			EkBomb_Color = {};
			EkBomb_OptionsFrame_SetDefaultColors();
		end
		EkBomb_OptionsFrame_SetColors();

		EkBomb_MessageFrame:SetClampedToScreen(false);
		EkBomb_MessageFrame_RestorePosition();
		EkBomb_MessageFrame:SetClampedToScreen(true);
		EkBomb_MessageFrame_SetOpacity();
		EkBomb_MessageFrame_SetBackground();

		-- Update predfined table using the user's saved settings.
		EkBomb_ListFrame_Bombs_Config();

		-- Sort the lists before we build the event list since
		-- the event list contains tables of subscript numbers.
		EkBomb_ListFrame_Sort(EkBomb_PredefinedTable1);
--!!		EkBomb_ListFrame_Sort(EkBomb_CustomTable1);
--!!		EkBomb_ListFrame_Sort(EkBomb_CustomTable2);

		-- Build the event list and register required events.
		EkBomb_EventList_Update();

		if (EkCheck_RegisterAddon) then
			EkCheck_RegisterAddon(EkBomb_EkCheck_Config);
		end

		if (EkBomb_Status == 1) then
			-- Enable warning checking.
			EkBomb_Enable();
		end

		if (EkBomb_ShowLink == 1) then
--!!			EkBomb_EventFrame_HookFunctions();
		end

		return;

	elseif (event == "PLAYER_DEAD") then
		if (EkBomb_FlashTimer or EkBomb_WarningTimer or EkBomb_MessageFrame:IsShown()) then
			EkBomb_Hide();
		end

		return;
	end
end

function EkBomb_Enable()
	-- ----------------
	-- Enable the warnings.
	-- ----------------
	EkBomb_Status = 1;

	EkBomb_Frame:RegisterEvent("PLAYER_DEAD");

	-- Register all events
	for k, v in pairs(EkBomb_EventList) do
		EkBomb_MessageFrame:RegisterEvent(k);
	end
end

function EkBomb_Disable()
	-- ----------------
	-- Disable the warnings.
	-- ----------------
	EkBomb_Status = 0;

	EkBomb_Frame:UnregisterEvent("PLAYER_DEAD");

	-- Unregister all events
	for k, v in pairs(EkBomb_EventList) do
		EkBomb_MessageFrame:UnregisterEvent(k);
	end
end

function EkBomb_Hide()
	-- ----------------
	-- Hide the warning window.
	-- ----------------
	EkBomb_MessageFrame:Hide();
end

function EkBomb_ShowBomb(aBomb, seconds)
	-- ----------------
	-- Show a specific bomb warning window.
	-- ----------------
	if (not seconds) then
		seconds = aBomb["timer"];
	end

	EkBomb_Show(aBomb["large"], aBomb["small"], seconds);
end

function EkBomb_Show(bigText, smallText, seconds)
	-- ----------------
	-- Show the specified text in the warning window.
	-- ----------------
	local fitScale;

	if (EkBomb_MessageFrame:IsShown()) then
		EkBomb_MessageFrame:Hide();
	end

	if (not bigText or type(bigText) ~= "string" or bigText == "") then
		bigText = " ";  -- Do not use 0 length string
	end

	if (not smallText or type(smallText) ~= "string" or smallText == "") then
		smallText = " ";  -- Do not use 0 length string
	end

	if (not seconds or type(seconds) ~= "number") then
		seconds = 30;
	end

	EkBomb_MessageFrame_SetOpacity();
	EkBomb_MessageFrame_SetBackground();

	EkBomb_MessageFrame_SetScale();  -- Reset to user scale (in case last scale was an override)

	EkBomb_MessageFrame_Message1:SetText(smallText);
	EkBomb_MessageFrame_Message2:SetText(bigText);

	-- Determine maximum scale to make warning fit on screen.
	-- (may not be exact due to the way the fontstring:GetStringWidth()
	-- varies with the scale of its frame)
	-- 
	fitScale = EkBomb_MessageFrame_CalcMaxScale();

	EkBomb_ScaleOverride = nil;  -- Default to using the user scale setting.
	-- If warning frame too big, then override the scale so it fits on screen.
	if (EkBomb_Scale/100 > fitScale) then
		EkBomb_MessageFrame_ScaleToFit();
		EkBomb_ScaleOverride = 1;
	else
		-- Set scale, adjust height, width.
		EkBomb_MessageFrame_SetScale();
	end

	-- Establish button width using the text that includes the timer
	EkBomb_MessageFrame_CloseButton:SetText(string.format(EkBomb_CloseButton_Format, seconds));
	EkBomb_MessageFrame_CloseButton:SetWidth(EkBomb_MessageFrame_CloseButton:GetTextWidth() + 50);

	if (EkBomb_ShowTimer == 0 or seconds <= 0) then
		EkBomb_MessageFrame_CloseButton:SetText(EkBomb_TEXT_Warning_Close); 
	end

	EkBomb_FlashTimer = 0;
	EkBomb_FlashState = 1;

	if (EkBomb_Sound == 1) then
		PlaySoundFile("Sound\\Doodad\\BellTollAlliance.wav");
	end

	if (seconds == 0) then
		EkBomb_WarningTimer = nil;  -- No automatic close timer.
	else
		EkBomb_WarningTimer = seconds;
	end

	EkBomb_MessageFrame:Show();
end

function EkBomb_ResetWindows()
	-- ----------------
	-- Reset position of all of the windows.
	-- ----------------
	EkBomb_MainFrame_Reset();
	EkBomb_MessageFrame_Reset();
	if (EkBomb_EventFrame) then
		EkBomb_EventFrame_Reset();
	end
end

function EkBomb_TestWarning(rand, listNum, sub, seconds, canDrag)
	-- ----------------
	-- Test the warning window.
	--
	-- Param: rand -- nil == show warnings sequentially
	--                1   == show random warning
	-- Param: listNum -- (optional) List number to display warning from (from 1 to #EkBomb_WarningLists).
	-- Param: sub -- (Only used if listNum specified). Subscript of item in the specified list.
	-- Param: seconds -- (optional) Force warning to use this number of seconds.
	-- Param: canDrag -- 1==User can drag warning window, 0==Cannot
	-- ----------------
	local list, item;
	local found;

	if (listNum) then
		-- Test using specific list.
		list = EkBomb_WarningLists[listNum];
		if (not list) then
			return;
		end
		if (#list == 0) then
			return;
		end
		if (sub) then
			-- Test using specific item in the list.
			if (sub <= 0) then
				sub = 1;
			elseif (sub > #list) then
				sub = #list;
			end
			item = list[sub];
			if (not item) then
				return;
			end
		else
			if (rand) then
				-- Pick a random warning.
				sub = random(1, #list);
			else
				-- Step through the warnings sequentially.
				sub = EkBomb_TestIndex + 1;
				if (sub <= 0 or sub > #list) then
					sub = 1;
				end
			end
		end
	else
		-- Test using all lists and items.

		-- Make sure at least one list has an item in it.
		found = nil;
		for i, list in ipairs(EkBomb_WarningLists) do
			if (#list > 0) then
				found = 1;
				break;
			end
		end
		if (not found) then
			return;
		end

		if (rand) then
			-- Pick a random warning.
			sub = 0;
			repeat
				listNum = random(1, #EkBomb_WarningLists);
				list = EkBomb_WarningLists[listNum];
				if (#list > 0) then
					sub = random(1, #list);
				else
					sub = 0;
				end
			until (sub ~= 0);
		else
			-- Step through the warnings sequentially.
			listNum = EkBomb_TestList;
			sub = EkBomb_TestIndex;
			if (sub < 0) then
				sub = 0;
			end
			repeat
				list = EkBomb_WarningLists[listNum];
				sub = sub + 1;
				if (sub > #list) then
					listNum = listNum + 1;
					if (listNum > #EkBomb_WarningLists) then
						listNum = 1;
					end
					sub = 0;
				end
			until (sub ~= 0);
		end
	end

	item = list[sub];

	if (canDrag == 1) then
		EkBomb_TestMode = 1;
	else
		EkBomb_TestMode = nil;
	end

	-- Show the (test) warning window.
	EkBomb_ShowBomb(item, seconds);

	EkBomb_TestMode = nil;

	-- Remember the last one we showed.
	EkBomb_TestList = listNum;
	EkBomb_TestIndex = sub;

	return;
end

function EkBomb_TestCommand()
	-- ----------------
	-- _test command
	-- ----------------
	if (EkBomb_TestFlag == 0) then
		EkBomb_TestFlag = 1;
		EkBomb_Print("Test mode enabled.");
	else
		EkBomb_TestFlag = 0;
		EkBomb_Print("Test mode disabled.");
	end

	if (not EkBomb_TestFlag_Warnings) then
		EkBomb_TestFlag_Warnings = 1;
		tinsert(EkBomb_PredefinedTable1,
			{
				-- Test warning (not localized):
				enabled = 1,
				name    = "Mob kill (Test)",
				small   = "Something was killed!",
				large   = "STOP!",
				timer   = 0,
				event   = "COMBAT_LOG_EVENT_UNFILTERED",

--				arg     = 1,
--				search  = EkBomb_SearchCode_Pattern,
--				value   = "^.+ dies",

				class   = "ANY",  -- Do not localize
				zone    = "",
				test    = "%s dies",
				id      = "mobkill",
				args    = {
					{
						arg    = COMBATLOG_ARG_event,
						search = EkBomb_SearchCode_Exact,
						value  = "UNIT_DIED",
					},
				}
			}
		)

		EkBomb_ListFrame_Sort(EkBomb_PredefinedTable1);
		EkBomb_EventList_Update();
		EkBomb_ListFrame_ListUpdate();
	end
end

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* Miscellaneous
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_Print(msg)
	-- ----------------
	-- Print a message.
	-- ----------------
	DEFAULT_CHAT_FRAME:AddMessage("<" .. EkBomb_Addon .. "> " .. msg);
	return;
end

function EkBomb_Print2(msg)
	-- ----------------
	-- Print a message.
	-- ----------------
	DEFAULT_CHAT_FRAME:AddMessage(msg);
	return;
end

function EkBomb_ReAnchor_Center_TopLeft(frame)
	-- ----------------
	-- Re-anchor the specified frame on its parent using a CENTER anchor point and a TOPLEFT relative point.
	-- ----------------
	local scale = frame:GetScale();
	local parent = frame:GetParent();

	local x = (frame:GetLeft() * scale) + ((frame:GetWidth() * scale)/ 2);
	local y = (parent:GetTop() - (frame:GetTop() * scale)) + ((frame:GetHeight() * scale) / 2);

	frame:ClearAllPoints();
	frame:SetPoint("CENTER", parent, "TOPLEFT", x / scale, -y / scale);
end

function EkBomb_ReAnchor_TopLeft_TopLeft(frame)
	-- ----------------
	-- Re-anchor the specified frame on its parent using a TOPLET anchor point and a TOPLEFT relative point.
	-- ----------------
	local scale = frame:GetScale();
	local parent = frame:GetParent();

	local x = (frame:GetLeft() * scale);
	local y = (parent:GetTop() - (frame:GetTop() * scale));

	frame:ClearAllPoints();
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", x / scale, -y / scale);
end

function EkBomb_ScrollBar_OnMouseWheel(scrollbar, direction)
	-- ----------
	-- User scrolled mouse wheel.
	-- ----------
	local sb = scrollbar;
	if (direction == 1) then
		-- Scrolled forward
		sb:SetValue(sb:GetValue() - (sb.ekheight / 2));
		PlaySound("UChatScrollButton");
	else
		-- Scrolled backward
		sb:SetValue(sb:GetValue() + (sb.ekheight / 2));
		PlaySound("UChatScrollButton");
	end
end

function EkBomb_SetTooltip_Cursor(owner, normalText)
	-- ----------
	-- Set tooltip anchored to the cursor.
	-- ----------
--	if (EkBomb_ShowTooltips ~= 1) then
--		return
--	end
	GameTooltip:SetOwner(owner, "ANCHOR_CURSOR");
	GameTooltip:SetText(normalText);
	GameTooltip:Show();
end

function EkBomb_SetTooltip_Anchor(owner, normalText, anchor, x, y, wrap)
	-- ----------
	-- Set tooltip using specified anchor.
	-- ----------
	local r, g, b, alpha;
--	if (EkBomb_ShowTooltips ~= 1) then
--		return
--	end
	GameTooltip:SetOwner(owner, anchor, x, y);
	GameTooltip:SetText(normalText, r, g, b, alpha, wrap);
	GameTooltip:Show();
end

function EkBomb_SetTooltip_Corner(owner, normalText, detailedText, anchor)
	-- ----------
	-- Set tooltip in lower right corner of screen.
	--
	-- Param: normalText -- Brief tooltip text.
	-- Param: detailedText -- Detailed tooltip text.
	-- ----------
	local r, g, b, alpha, wrap;
	if (EkBomb_ShowTooltips ~= 1) then
		return
	end
	r = 1;
	g = 1;
	b = 1;
	alpha = 1;
	wrap = 1;
	if (EkBomb_ShowAtCursor == 1) then
		if (not anchor) then
			local oType = owner:GetObjectType();
			if (oType == "CheckButton") then
				anchor = "ANCHOR_LEFT";
			elseif (oType == "Button") then
				anchor = "ANCHOR_BOTTOMRIGHT";
			elseif (oType == "EditBox") then
				anchor = "ANCHOR_RIGHT";
			elseif (oType == "Frame") then
				anchor = "ANCHOR_RIGHT";
			else
				anchor = "ANCHOR_RIGHT";
			end
		end
		GameTooltip:SetOwner(owner, anchor);
	else
		GameTooltip_SetDefaultAnchor(GameTooltip, owner);
	end
	if (normalText) then
		GameTooltip:SetText(normalText, r, g, b, alpha, wrap);
		if (detailedText) then
			GameTooltip:AddLine(detailedText, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1);
		end
	else
		GameTooltip:SetText(detailedText, r, g, b, alpha, wrap);
	end
	GameTooltip:Show();
end

function EkBomb_HideTooltip()
	-- ----------
	-- Hide the tooltip.
	-- ----------
	GameTooltip:Hide();
end

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb_MessageFrame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_MessageFrame_OnLoad()
	-- ----------------
	-- Message frame loaded.
	-- ----------------
	EkBomb_MessageFrame:Hide();

	EkBomb_MessageFrame_Message1:SetTextHeight(32);
	EkBomb_MessageFrame_Message2:SetTextHeight(200);
end

function EkBomb_MessageFrame_OnShow()
	-- ----------------
	-- Message frame being shown.
	-- ----------------
	EkBomb_MessageFrame_CloseButton:Show();

	if (EkBomb_TestMode) then
		-- User pressed the "Test" button to display the warning frame.
		-- Allow the frame to be dragged.
		EkBomb_MessageFrame_SetMovable(1);
	end
end

function EkBomb_MessageFrame_OnHide()
	-- ----------------
	-- Message frame being hidden.
	-- ----------------
	EkBomb_MessageFrame_CloseButton:Hide();

	EkBomb_MessageFrame:StopMovingOrSizing();

	EkBomb_MessageFrame_SetMovable(0);

	EkBomb_WarningTimer = nil;
	EkBomb_FlashTimer = nil;
	EkBomb_MessageFrame:SetAlpha(0);
end

function EkBomb_MessageFrame_SetMovable(enabled)
	-- ----------------
	-- Set the movable state of the window.
	-- ----------------
	if (enabled == 1) then
		EkBomb_MessageFrame:SetMovable(true);
		EkBomb_MessageFrame_Drag:Show();
	else
		EkBomb_MessageFrame_Drag:Hide();
		EkBomb_MessageFrame:SetMovable(false);
	end
end

function EkBomb_MessageFrame_OnEvent(self, event, ...)
	-- ----------------
	-- Message frame event handler.
	-- ----------------
	if (EkBomb_Status == 0) then
		return;
	end

	EkBomb_MessageFrame_Match(event, ...);
end

function EkBomb_MessageFrame_Match(eventName, ...)
	-- ----------
	-- Try and match the event to one in our tables.
	-- ----------
	local eventItem, eventList, list, item;
	local arg, name;
	local start, pos2, found;
	local class1, class2;
	local zone1, zone2, zone3, zone4;
	local args, argItem;

	-- Don't show warning if player is dead.
	if (UnitIsDeadOrGhost("player")) then
		return;
	end

	-- Are there any warnings using this event name?
	eventItem = EkBomb_EventList[eventName];
	if (not eventItem) then
		return;
	end

	zone1 = GetRealZoneText();
	zone2 = GetZoneText();
	zone3 = GetSubZoneText();
	zone4 = GetMinimapZoneText();

	-- Scan the warning lists (custom character, custom general, predefined general).
	for i = 1, #EkBomb_WarningLists do
		-- Get the event list associated with this warning list.
		eventList = eventItem[i];
		-- Any events associated with this list?
		if (eventList) then
			-- Get the warning list
			list = EkBomb_WarningLists[i];
			-- Process the subscript numbers found in this event list.
			for j, sub in ipairs(eventList) do
				-- Get the warning item from the warning list.
				item = list[sub];

				found = 1;
				name = nil;

				-- Test for class (if applicable)
				if (found) then
					if (item["class"] ~= "ANY") then
						-- There is a class restriction on this bomb.
						class1, class2 = UnitClass("player");
						if (item["class"] ~= class2) then
							found = nil;
						end
					end
				end

				-- Test for zone (if applicable)
				if (found) then
					if (item["zone"] ~= "") then
						if (item["zone"] == zone1) then
						elseif (item["zone"] == zone2) then
						elseif (item["zone"] == zone3) then
						elseif (item["zone"] == zone4) then
						else
							found = nil;
						end
					end
				end

				-- If user wants to test an argument value...
				if (found) then
					if (not EkBomb_MessageFrame_MatchArgs(eventName, item, ...)) then
						found = nil;
					end
				end

				-- If we passed all the tests, then show the warning.
				if (found) then
					EkBomb_ShowBomb(item);
					break;
				end
			end

			if (found) then
				break;
			end
		end
	end
end

function EkBomb_MessageFrame_MatchArgs(eventName, item, ...)
	-- ----------------
	-- Match arguments
	-- ----------------
	local argItem, arg, args, argType;
	local found;
	local start, pos2, name;

	found = 1;
	args = item["args"];

	if (not args) then
		-- No arguments, so it matches the event.
		return 1;
	end

	for i = 1, #args do
		argItem = args[i];
		if (argItem["arg"] > 0) then

			-- Get the argument value
			arg = select(argItem["arg"], ...);
			argType = type(arg);

			if (argItem["search"] == EkBomb_SearchCode_Exact) then
				if (argType == "string" or argType == "number") then
					-- Test for an exact match.
					if (arg ~= argItem["value"]) then
						found = nil;
					end
				else
					-- Unsupported argument type
					found = nil;
				end

			elseif (argItem["search"] == EkBomb_SearchCode_Pattern) then
				if (argType == "string" or argType == "number") then
					-- Test for a pattern match, and try to capture a player's name.
					start, pos2, name = string.find(arg, argItem["value"]);
					if (not start) then
						found = nil;
					else
						-- Captured name must match the current player's name.
						if (name) then
							if (name ~= UnitName("player")) then
								found = nil;
							end
						end
					end
				else
					-- Unsupported argument type
					found = nil;
				end

			elseif (argItem["search"] == EkBomb_SearchCode_Filter) then
				if (argType == "string" or argType == "number") then
					-- Test if combat log flag value passes a filter.
					local filter = EkBomb_FilterBits[argItem["value"]];
					if (filter) then
						if (not CombatLog_Object_IsA(arg, filter)) then
							found = nil;
						end
					end
				else
					-- Unsupported argument type
					found = nil;
				end

			else
				-- Unknown search code
			end
		end
	end

	return found;
end

function EkBomb_MessageFrame_OnUpdate(self, elapsed)
	-- ----------------
	-- Update routine for Warning Frame.
	-- ----------------
	if (EkBomb_WarningTimer) then
		if (not EkBomb_PauseWarningTimer) then
			-- Count down the warning timer. When done, hide the warning window.
			EkBomb_WarningTimer = EkBomb_WarningTimer - elapsed;
			EkBomb_WarningTimer2 = EkBomb_WarningTimer2 - elapsed;

			-- Set appropriate text for close button (user may toggle while warning on screen).
			if (EkBomb_WarningTimer2 <= 0) then
				EkBomb_WarningTimer2 = 0.1;
				if (EkBomb_ShowTimer == 1) then
					EkBomb_MessageFrame_CloseButton:SetText(string.format(EkBomb_CloseButton_Format, ceil(EkBomb_WarningTimer)));
				else
					EkBomb_MessageFrame_CloseButton:SetText(EkBomb_TEXT_Warning_Close); 
				end
			end

			if (EkBomb_WarningTimer < 0.1) then
				EkBomb_Hide();
			end
		end
	end

	if (EkBomb_FlashTimer) then
		-- Count down the flash timer. When done, "flash" the window and restart flash timer.
		EkBomb_FlashTimer = EkBomb_FlashTimer - elapsed;
		if (EkBomb_FlashTimer < 0.1) then
			EkBomb_MessageFrame_SetOpacity();
			if (EkBomb_FlashState == 1) then
				EkBomb_FlashTimer = 2;
				EkBomb_FlashState = 2;

			elseif (EkBomb_FlashWindow == 1) then
				EkBomb_FlashTimer = 1;
				EkBomb_FlashState = 1;
			end
		end
	end
end

function EkBomb_MessageFrame_Reset()
	-- ----------------
	-- Reset position of the warning window.
	-- ----------------
	EkBomb_MessageFrame:SetScale(EkBomb_Scale / 100);
	EkBomb_MessageFrame:ClearAllPoints();
	EkBomb_MessageFrame:SetPoint("CENTER", "UIParent", "CENTER");
	EkBomb_MessageFrame_SavePosition();
end

function EkBomb_MessageFrame_SavePosition()
	-- ----------------
	-- Save current position of the warning frame.
	-- ----------------
	local ap, rf, rp, rx, ry = EkBomb_MessageFrame:GetPoint();
	if (rf) then
		rf = rf:GetName();
	end
	if (type(EkBomb_WarningPosition) ~= "table") then
		EkBomb_WarningPosition = {};
	end
	EkBomb_WarningPosition.ap = ap;
	EkBomb_WarningPosition.rf = rf;
	EkBomb_WarningPosition.rp = rp;
	EkBomb_WarningPosition.rx = rx;
	EkBomb_WarningPosition.ry = ry;
	EkBomb_WarningPosition.sc = EkBomb_MessageFrame:GetScale();
end

function EkBomb_MessageFrame_RestorePosition()
	-- ----------------
	-- Restore current position of the warning frame.
	-- ----------------
	local pos = EkBomb_WarningPosition;
	if (pos) then
		EkBomb_MessageFrame:SetScale(pos.sc);
		EkBomb_MessageFrame:ClearAllPoints();
		EkBomb_MessageFrame:SetPoint(pos.ap, pos.rf, pos.rp, pos.rx, pos.ry);
	else
		EkBomb_MessageFrame_Reset();
	end
end

function EkBomb_MessageFrame_SetScale(scale)
	-- ----------------
	-- Change the warning frame's scale.
	-- ----------------
	if (not scale) then
		scale = EkBomb_Scale / 100;
	end

	EkBomb_MessageFrame:SetClampedToScreen(false);

	EkBomb_ReAnchor_Center_TopLeft(EkBomb_MessageFrame);
	EkBomb_MessageFrame_SetScale2(scale);

	-- Adjust frame width and height
	local width = EkBomb_MessageFrame_CalcWidth();
	local height = EkBomb_MessageFrame_CalcHeight();

	EkBomb_MessageFrame:SetWidth(width);
	EkBomb_MessageFrame_Drag:SetWidth(width);

	EkBomb_MessageFrame:SetHeight(height);
	EkBomb_MessageFrame_Drag:SetHeight(height);

	EkBomb_MessageFrame:SetClampedToScreen(true);
end

function EkBomb_MessageFrame_SetScale2(scale)
	-- ----------------
	-- Change a window's scale without appearing to change
	-- its position while maintaining the current anchor point.
	-- ----------------
	if (not scale) then
		scale = EkBomb_Scale / 100;
	end

	local frame = EkBomb_MessageFrame;
	local parent = frame:GetParent();

	local ap, rf, rp, rx, ry = frame:GetPoint();

	local oldScale = frame:GetScale();
	local newScale = scale;

	local frameX, frameY, parentX, parentY;
	local val;

	-- Determine frame X and Y values.
	if (ap == "TOPLEFT" or ap == "BOTTOMLEFT" or ap == "LEFT") then
		frameX = frame:GetLeft();
	elseif (ap == "TOPRIGHT" or ap == "BOTTOMRIGHT" or ap == "RIGHT") then
		frameX = frame:GetRight();
	elseif (ap == "TOP" or ap == "BOTTOM" or ap == "CENTER") then
		frameX = frame:GetCenter();
	end

	if (ap == "TOPLEFT" or ap == "TOPRIGHT" or ap == "TOP") then
		frameY = frame:GetTop();
	elseif (ap == "BOTTOMLEFT" or ap == "BOTTOMRIGHT" or ap == "BOTTOM") then
		frameY = frame:GetBottom();
	elseif (ap == "LEFT" or ap == "RIGHT" or ap == "CENTER") then
		val, frameY = frame:GetCenter();
	end

	-- Determine parent X and Y values.
	if (rp == "TOPLEFT" or rp == "BOTTOMLEFT" or rp == "LEFT") then
		parentX = parent:GetLeft();
	elseif (rp == "TOPRIGHT" or rp == "BOTTOMRIGHT" or rp == "RIGHT") then
		parentX = parent:GetRight();
	elseif (rp == "TOP" or rp == "BOTTOM" or rp == "CENTER") then
		parentX = parent:GetCenter();
	end

	if (rp == "TOPLEFT" or rp == "TOPRIGHT" or rp == "TOP") then
		parentY = parent:GetTop();
	elseif (rp == "BOTTOMLEFT" or rp == "BOTTOMRIGHT" or rp == "BOTTOM") then
		parentY = parent:GetBottom();
	elseif (rp == "LEFT" or rp == "RIGHT" or rp == "CENTER") then
		val, parentY = parent:GetCenter();
	end

	-- Calculate offsets
	local yOffset = frameY - (parentY / oldScale);
	local xOffset = frameX - (parentX / oldScale);

	-- Calculate the new X and Y offsets for the new scale.
	local newX = xOffset * (oldScale / newScale);
	local newY = yOffset * (oldScale / newScale);

	frame:SetScale(newScale);
	frame:ClearAllPoints();
	frame:SetPoint(ap, rf, rp, newX, newY);
end

function EkBomb_MessageFrame_ScaleToFit()
	-- ----------
	-- Calculate a scale that will fit the message frame on screen.
	--
	-- Problem: fontstring:GetStringWidth() changes as the frame scale changes.
	-- Loop until we get a scale that results in a frame that fits on screen.
	-- Stop if we've done a number of passes without finding one.
	-- ----------
	local scale, fitScale, width, uWidth;
	local count;

	count = 0;
	uWidth = UIParent:GetWidth();

	repeat
		count = count + 1;

		fitScale = EkBomb_MessageFrame_CalcMaxScale();
		EkBomb_MessageFrame_SetScale(fitScale);

		scale = EkBomb_MessageFrame:GetScale();
		width = EkBomb_MessageFrame_CalcWidth() * scale;

		-- EkBomb_Print(count .. ", scale=" .. scale .. ", MW=" .. width);

		if (width <= uWidth) then
			-- Fits on screen.
			break;
		end
	until (count >= 19)
end

function EkBomb_MessageFrame_CalcHeight()
	-- ----------
	-- Calculate height needed for the message frame.
	-- ----------
	return 280;
end

function EkBomb_MessageFrame_CalcWidth()
	-- ----------
	-- Calculate width needed for the message frame.
	-- ----------

	-- Determine the wider of the two message strings.
	local s1Width = EkBomb_MessageFrame_Message1:GetStringWidth();
	local s2Width = EkBomb_MessageFrame_Message2:GetStringWidth();
	local width;

	if (s1Width > s2Width) then
		width = s1Width;
	else
		width = s2Width;
	end

	if (width < 200) then
		width = 200;
	end

	return width;
end

function EkBomb_MessageFrame_CalcMaxScale()
	-- ----------
	-- Calculate approximate maximum scale to fit all of the warning on screen.
	-- ----------
	local maxScale;

	local width = EkBomb_MessageFrame_CalcWidth();
	local height = EkBomb_MessageFrame_CalcHeight();

	local horScale = UIParent:GetWidth() / width;
	local verScale = UIParent:GetHeight() / height;

	if (horScale > verScale) then
		maxScale = verScale;
	else
		maxScale = horScale;
	end

	if (maxScale < 0.1) then
		maxScale = 0.1;
	end

	return maxScale;
end

function EkBomb_MessageFrame_SetBackground()
	-- ----------------
	-- Show/hide the warning frame's background.
	-- ----------------
	if (EkBomb_Background == 1) then
		EkBomb_MessageFrame:EnableDrawLayer("BACKGROUND");
	else
		EkBomb_MessageFrame:DisableDrawLayer("BACKGROUND");
	end
end

function EkBomb_MessageFrame_SetOpacity()
	-- ----------------
	-- Set the warning frame's opacity.
	-- ----------------
	local opacity;

	if (EkBomb_FlashState == 1 or EkBomb_FlashWindow == 0) then
		opacity = EkBomb_Opacity;
	else
		opacity = EkBomb_Opacity - 20;
		if (opacity <  1)
			then opacity = 1;
		end
	end

	EkBomb_MessageFrame:SetAlpha(opacity / 100);
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb_MainFrame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_MainFrame_OnLoad(self)
	-- ----------
	-- Frame has been loaded.
	-- ----------
	local frame = EkBomb_MainFrame;

--	frame:SetClampedToScreen(1);

	tinsert(UISpecialFrames, "EkBomb_MainFrame");

	PanelTemplates_SetNumTabs(self, #EkBomb_MainFrame_SUBFRAMES);

	-- Initial tab to be selected
	frame.selectedTab = EkBomb_MainFrame_TabOptions;

	PanelTemplates_UpdateTabs(self);
end

function EkBomb_MainFrame_OnShow()
	-- ----------
	-- Show the frame.
	-- ----------
	local fname = "EkBomb_MainFrame";
	local frame = _G[fname];
	local tab = frame.selectedTab;

	fname = EkBomb_MainFrame_SUBFRAMES[tab];

	EkBomb_MainFrame_ShowSubFrame(fname);
	PlaySound("igMainMenuOpen");

	EkBomb_MainFrame_Title:SetText(EkBomb_Addon .. " " .. EkBomb_Version);
end

function EkBomb_MainFrame_OnHide()
	-- ----------
	-- Hide the frame.
	-- ----------
	PlaySound("igMainMenuClose");
	EkBomb_MainFrame_HideSubFrames();
	EkBomb_MainFrame_StopMoving();
end

function EkBomb_MainFrame_StopMoving()
	-- ----------
	-- Stop moving the frame.
	-- ----------
	EkBomb_MainFrame:StopMovingOrSizing();
end

function EkBomb_MainFrame_Refresh()
	-- ----------
	-- Refresh the frame.
	-- ----------
	local frame = EkBomb_MainFrame;
	local tab = frame.selectedTab;

	if ( tab == EkBomb_MainFrame_TabOptions ) then
		EkBomb_OptionsFrame_OnShow();

	elseif ( tab == EkBomb_MainFrame_TabPredefined ) then
		EkBomb_PredefinedFrame_OnShow();

	elseif ( tab == EkBomb_MainFrame_TabCustom ) then
		EkBomb_CustomFrame_OnShow();

	end
end

function EkBomb_MainFrame_ShowSubFrame(frameName)
	-- ----------
	-- Show/hide the sub frames.
	-- ----------
	EkBomb_MainFrame_ShowSubFrameTable(frameName, EkBomb_MainFrame_SUBFRAMES);
end

function EkBomb_MainFrame_ShowSubFrameTable(frameName, frameNameTable)
	-- ----------
	-- Show/hide the sub frame names listed in a table.
	-- ----------
	local frame;
	EkBomb_MainFrame_HideSubFrames();
	for index, value in pairs(frameNameTable) do
		if ( type(value) == "table" ) then
			EkBomb_MainFrame_ShowSubFrameTable(frameName, value);
		else
			frame = _G[value];
			if ( frame ) then
				if ( value == frameName ) then
					frame:Show();
				else
					frame:Hide();
				end
			end
		end
	end
end

function EkBomb_MainFrame_HideSubFrames()
	-- ----------
	-- Hide the sub frames.
	-- ----------
	EkBomb_MainFrame_HideSubFrameTable(EkBomb_MainFrame_SUBFRAMES);
end

function EkBomb_MainFrame_HideSubFrameTable(frameNameTable)
	-- ----------
	-- Hide the sub frame names listed in a table.
	-- ----------
	local frame;
	for index, value in pairs(frameNameTable) do
		if ( type(value) == "table" ) then
			EkBomb_MainFrame_HideSubFrameTable(value);
		else
			frame = _G[value];
			if ( frame ) then
				frame:Hide();
			end
		end
	end
end

function EkBomb_MainFrame_GetTab()
	-- ----------
	-- Get current tab number.
	-- ----------
	return EkBomb_MainFrame.selectedTab;
end

function EkBomb_MainFrame_Tab_OnClick(self)
	-- ----------
	-- User clicked a tab on the frame.
	-- ----------
	PanelTemplates_Tab_OnClick(self, EkBomb_MainFrame);
	EkBomb_MainFrame_OnShow();
end

function EkBomb_MainFrame_GotoTab(tab)
	-- ----------
	-- Display one of the UI tab pages.
	-- ----------
	PanelTemplates_SetTab(EkBomb_MainFrame, tab);
	EkBomb_MainFrame_OnShow();
end

function EkBomb_MainFrame_Reset()
	-- ----------------
	-- Reset position of the main window.
	-- ----------------
	EkBomb_MainFrame:ClearAllPoints();
	EkBomb_MainFrame:SetPoint("CENTER", "UIParent", "CENTER");
end

function EkBomb_MainFrame_TestButton_OnClick()
	-- ----------------
	-- User clicked the test button.
	-- ----------------
	local rand;

	EkBomb_Hide();
	rand = IsControlKeyDown();
	EkBomb_TestWarning(rand, nil, nil, 0, 1);
end

function EkBomb_MainFrame_CloseButton_OnClick()
	-- ----------------
	-- User clicked the close button.
	-- ----------------
	EkBomb_MainFrame:Hide();
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb_OptionsFrame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_OptionsFrame_OnLoad(self)
	-- ----------------
	-- Options frame is being loaded.
	-- ----------------
	self:SetBackdropBorderColor(0.4, 0.4, 0.4);
	self:SetBackdropColor(0.10, 0.10, 0.10);

	EkBomb_OptionsFrame_Create_BufferSize_Items();
end

function EkBomb_OptionsFrame_OnShow()
	-- ----------------
	-- Options frame is being shown.
	-- ----------------
	EkBomb_OptionsFrame_SetColors();

	EkBomb_OptionsFrame_FlashCB:SetChecked(EkBomb_FlashWindow);
	EkBomb_OptionsFrame_BackgroundCB:SetChecked(EkBomb_Background);
	EkBomb_OptionsFrame_SoundCB:SetChecked(EkBomb_Sound)
	EkBomb_OptionsFrame_ShowTimerCB:SetChecked(EkBomb_ShowTimer);
	EkBomb_OptionsFrame_ShowTooltipsCB:SetChecked(EkBomb_ShowTooltips);
	EkBomb_OptionsFrame_ShowAtCursorCB:SetChecked(EkBomb_ShowAtCursor);
	EkBomb_OptionsFrame_ShowAtCursorCB_Update();

--[[ !!
	EkBomb_OptionsFrame_ShowLinkCB:SetChecked(EkBomb_ShowLink);
	EkBomb_OptionsFrame_BufferSizeEB:SetNumber(EkBomb_EventBufferSize);
	EkBomb_OptionsFrame_BufferSizeEB_Refresh();
]]

	EkBomb_OptionsFrame_EnableCB:SetChecked(EkBomb_Status);
end

function EkBomb_OptionsFrame_Scale_OnShow(self)
	-- ----------------
	-- Scale slider is being shown.
	-- ----------------
	local slider = self;
	_G[slider:GetName() .. "High"]:SetText("300%");
	_G[slider:GetName() .. "Low"]:SetText("25%");
	_G[slider:GetName() .. "Text"]:SetText(EkBomb_TEXT_Config_Scale .. " = " .. floor(EkBomb_Scale + 0.5) .. "%");

	slider:SetMinMaxValues(0.25, 3);
	slider:SetValueStep(0.01);

	EkBomb_ScaleOnShow = 1;  -- Set flag to tell EkBomb_OptionsFrame_Scale_OnValueChanged() that this routine changed the slider's value.
	slider:SetValue(EkBomb_Scale / 100);
end

function EkBomb_OptionsFrame_Scale_OnValueChanged(self)
	-- ----------------
	-- Value of scale slider has changed.
	-- ----------------
	EkBomb_Scale = floor(self:GetValue() * 100 + 0.5);
	_G[self:GetName() .. "Text"]:SetText(EkBomb_TEXT_Config_Scale .. " = " .. EkBomb_Scale .. "%");

	-- Set the scale of the warning frame.
	if (EkBomb_ScaleOnShow) then
		-- EkBomb_OptionsFrame_Scale_OnShow() changed the value of the slider.
		EkBomb_ScaleOnShow = nil;
		-- If we are currently overriding the user scale setting...
		if (EkBomb_ScaleOverride) then
			-- Don't change the warning frame's scale this time.
			-- This is to stop the warning frame scale from changing if the user
			-- opens the EkBomb options window containing the slider while there is
			-- a warning window display that is using an override scale.
			-- Once the options window is open they are free to change the slider
			-- and it will effectively override the override.
			return;
		end
	end
	EkBomb_MessageFrame_SetScale();
end

function EkBomb_OptionsFrame_Opacity_OnShow(self)
	-- ----------------
	-- Opacity slider is being shown.
	-- ----------------
	local slider = self;
	_G[slider:GetName() .. "High"]:SetText("100%");
	_G[slider:GetName() .. "Low"]:SetText("0%");
	_G[slider:GetName() .. "Text"]:SetText(EkBomb_TEXT_Config_Opacity .. " = " .. floor(EkBomb_Opacity + 0.5) .. "%");

	slider:SetMinMaxValues(0, 1);
	slider:SetValueStep(0.01);
	slider:SetValue(EkBomb_Opacity / 100);
end

function EkBomb_OptionsFrame_Opacity_OnValueChanged(self)
	-- ----------------
	-- Value of opacity slider has changed.
	-- ----------------
	EkBomb_Opacity = floor(self:GetValue() * 100 + 0.5);
	_G[self:GetName() .. "Text"]:SetText(EkBomb_TEXT_Config_Opacity .. " = " .. EkBomb_Opacity .. "%");
	EkBomb_MessageFrame_SetOpacity();
end

function EkBomb_OptionsFrame_ColorPicker(key)
	-- ----------
	-- User wants to pick a color.
	-- ----------
	local frame = EkBomb_ColorPickerFrame;
	local color = EkBomb_Color[key];
	local defcolor = EkBomb_DefaultColor[key];

	if (not color) then
		color = { r = defcolor.r, g = defcolor.g, b = defcolor.b };
		EkBomb_Color[key] = color;
	end

	frame.key = key;

	frame.r = color.r;
	frame.g = color.g;
	frame.b = color.b;

	frame.original = {r = color.r, g = color.g, b = color.b};

	frame.func = EkBomb_OptionsFrame_ChangeColor;
	frame.cancelFunc = EkBomb_OptionsFrame_CancelColor;
	frame.defaultFunc = EkBomb_OptionsFrame_DefaultColor;

	EkBomb_ColorPickerFrameTitle:SetText(EkBomb_TEXT_Config_SelectColor);
	EkBomb_ColorPickerFrame:SetColorRGB(color.r, color.g, color.b);
	EkBomb_ColorPickerFrame:Show();
end

function EkBomb_OptionsFrame_ChangeColor()
	-- ----------
	-- User has selected a color.
	-- ----------
	local r, g, b = EkBomb_ColorPickerFrame:GetColorRGB();
	local key = EkBomb_ColorPickerFrame.key;
	local color = EkBomb_Color[key];

	color.r = r;
	color.g = g;
	color.b = b;

	EkBomb_OptionsFrame_SetColors();
end

function EkBomb_OptionsFrame_CancelColor()
	-- ----------
	-- User has cancelled the color selection.
	-- ----------
	local key = EkBomb_ColorPickerFrame.key;
	local original = EkBomb_ColorPickerFrame.original;
	local color = EkBomb_Color[key];

	color.r = original.r;
	color.g = original.g;
	color.b = original.b;

	EkBomb_OptionsFrame_SetColors();
end

function EkBomb_OptionsFrame_DefaultColor()
	-- ----------
	-- User has chosen the default color selection.
	-- ----------
	local frame = EkBomb_ColorPickerFrame;
	local key = EkBomb_ColorPickerFrame.key;
	local defcolor = EkBomb_DefaultColor[key];
	local color = EkBomb_Color[key];

	color.r = defcolor.r;
	color.g = defcolor.g;
	color.b = defcolor.b;

	frame.r = color.r;
	frame.g = color.g;
	frame.b = color.b;

	EkBomb_ColorPickerFrame:SetColorRGB(color.r, color.g, color.b);

	EkBomb_OptionsFrame_SetColors();
end

function EkBomb_OptionsFrame_SetColors()
	-- ----------
	-- Set the colors.
	-- ----------
	local color;

	color = EkBomb_Color["small"];
	EkBomb_MessageFrame_Message1:SetTextColor(color.r, color.g, color.b);
	EkBomb_OptionsFrame_Opacity_SmallColorSwatchNormalTexture:SetVertexColor(color.r, color.g, color.b);

	color = EkBomb_Color["large"];
	EkBomb_MessageFrame_Message2:SetTextColor(color.r, color.g, color.b);
	EkBomb_OptionsFrame_Opacity_LargeColorSwatchNormalTexture:SetVertexColor(color.r, color.g, color.b);

	color = EkBomb_Color["back"];
	EkBomb_MessageFrame_Background:SetVertexColor(color.r, color.g, color.b);
	EkBomb_OptionsFrame_Opacity_BackgroundColorSwatchNormalTexture:SetVertexColor(color.r, color.g, color.b);
end

function EkBomb_OptionsFrame_SetDefaultColors()
	-- ----------------
	-- Reset to default colors.
	-- ----------------
	for k, v in pairs(EkBomb_DefaultColor) do
		local color;
		color = {r = v.r, g = v.g, b = v.b};
		EkBomb_Color[k] = color;
	end
end

function EkBomb_OptionsFrame_FlashCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'periodically blink the window' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_FlashWindow = 1;
	else
		EkBomb_FlashWindow = 0;
	end
end

function EkBomb_OptionsFrame_FlashCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'periodically blink the window' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_Flash_Tooltip1, EkBomb_TEXT_Config_Flash_Tooltip2);
end

function EkBomb_OptionsFrame_BackgroundCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'show warning window background' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_Background = 1;
	else
		EkBomb_Background = 0;
	end
	EkBomb_MessageFrame_SetBackground();
end

function EkBomb_OptionsFrame_BackgroundCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'show warning window background' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_Background_Tooltip1, EkBomb_TEXT_Config_Background_Tooltip2);
end

function EkBomb_OptionsFrame_SoundCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'play a sound' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_Sound = 1;
	else
		EkBomb_Sound = 0;
	end
end

function EkBomb_OptionsFrame_SoundCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'play a sound' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_Sound_Tooltip1, EkBomb_TEXT_Config_Sound_Tooltip2);
end

function EkBomb_OptionsFrame_ShowTimerCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'show timer' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_ShowTimer = 1;
	else
		EkBomb_ShowTimer = 0;
	end
end

function EkBomb_OptionsFrame_ShowTimerCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'show timer' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_ShowTimer_Tooltip1, EkBomb_TEXT_Config_ShowTimer_Tooltip2);
end

function EkBomb_OptionsFrame_ShowTooltipsCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'show detailed tooltips' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_ShowTooltips = 1;
	else
		EkBomb_ShowTooltips = 0;
	end
	EkBomb_OptionsFrame_ShowAtCursorCB_Update();
end

function EkBomb_OptionsFrame_ShowTooltipsCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'Show detailed tooltips' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_ShowTooltips_Tooltip1, EkBomb_TEXT_Config_ShowTooltips_Tooltip2);
end

function EkBomb_OptionsFrame_ShowAtCursorCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'show detailed tooltips at cursor' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_ShowAtCursor = 1;
	else
		EkBomb_ShowAtCursor = 0;
	end
end

function EkBomb_OptionsFrame_ShowAtCursorCB_Update()
	-- ----------------
	-- Update the show tooltips at cursor checkboxes.
	-- ----------------
	if (EkBomb_ShowTooltips == 1) then
		EkBomb_OptionsFrame_ShowAtCursorCB:Enable();
	else
		EkBomb_OptionsFrame_ShowAtCursorCB:Disable();
	end
end

function EkBomb_OptionsFrame_ShowAtCursorCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'Show detailed tooltips at cursor' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_ShowAtCursor_Tooltip1, EkBomb_TEXT_Config_ShowAtCursor_Tooltip2);
end

function EkBomb_OptionsFrame_ShowLinkCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'show event link' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_ShowLink = 1;
		EkBomb_EventFrame_HookFunctions();
	else
		EkBomb_ShowLink = 0;
	end
	EkBomb_OptionsFrame_BufferSizeEB_Refresh();
end

function EkBomb_OptionsFrame_ShowLinkCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'show event link' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_ShowLink_Tooltip1, EkBomb_TEXT_Config_ShowLink_Tooltip2);
end

function EkBomb_OptionsFrame_Create_BufferSize_Items()
	-- ----------
	-- Create items to edit buffer size.
	-- ----------
	local fs, eb, b;
	local frame = EkBomb_OptionsFrame;
	local fname = frame:GetName();

	-- Message text
	frame:CreateFontString("EkBomb_OptionsFrame_BufferSizeText", "ARTWORK", "GameFontHighlight");
	fs = EkBomb_OptionsFrame_BufferSizeText;
	fs:SetHeight(10);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("TOPLEFT", "EkBomb_OptionsFrame_ShowLinkCB", "BOTTOMLEFT", 28, -5);
	fs:SetText(EkBomb_TEXT_Config_BufferSize_Text);
	fs:Hide();

	-- Buffer size edit box
	eb = CreateFrame("EditBox", "EkBomb_OptionsFrame_BufferSizeEB", frame, "EkBomb_EditBox_Template");
	eb:SetPoint("LEFT", "EkBomb_OptionsFrame_BufferSizeText", "RIGHT", 15, 0);
	eb:SetNumeric(1);
	eb:SetMaxLetters(5);  -- 7 chat frames times 128 lines per frame = 896. But use 5 digits to allow more than the number of chat lines.
	eb:SetWidth(50);
	eb:SetNumber(1);
	eb.ekTooltip1 = "EkBomb_TEXT_Config_BufferSize_Tooltip1";
	eb.ekTooltip2 = "EkBomb_TEXT_Config_BufferSize_Tooltip2";
	eb.ekEditTextChanged = EkBomb_OptionsFrame_BufferSizeEB_TextChanged;
	eb.ekEditStop = EkBomb_OptionsFrame_BufferSizeEB_FocusLost;
	eb:SetScript("OnEnterPressed", EkBomb_OptionsFrame_SetNewSize);
	eb:Hide();

	-- 'Set' button
	b = CreateFrame("Button", "EkBomb_OptionsFrame_BufferSize_Button", frame, "UIPanelButtonTemplate");
	b:SetHeight(30);
	b:SetWidth(60);
	b:SetPoint("LEFT", "EkBomb_OptionsFrame_BufferSizeEB", "RIGHT", 15, 0);
	b:SetText(EkBomb_TEXT_Config_BufferSize_SetButton);
	b:Disable();
	b:SetScript("OnClick", EkBomb_OptionsFrame_SetNewSize);
	b:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_BufferSize_SetButton_Tooltip1, EkBomb_TEXT_Config_BufferSize_SetButton_Tooltip2); end );
	b:SetScript("OnLeave", function() EkBomb_HideTooltip(); end );
	b:Hide();
end

function EkBomb_OptionsFrame_SetNewSize()
	-- ----------
	-- Set new size of event buffer.
	-- ----------
	local eb = EkBomb_OptionsFrame_BufferSizeEB;
	local bu = EkBomb_OptionsFrame_BufferSize_Button;
	local newSize;
	newSize = eb:GetNumber();
	if (newSize < 1) then
		newSize = 1;
	end
	EkBomb_EventFrame_ResizeBuffer(newSize);
	bu:Hide();
	eb:ClearFocus();
end

function EkBomb_OptionsFrame_BufferSizeEB_TextChanged(self)
	-- ----------
	-- Text inside the buffer size editbox has changed.
	-- ----------
	local bu = EkBomb_OptionsFrame_BufferSize_Button;
	local val;

	-- If this editbox doesn't have focus...
	if (not self.ekEditing) then
		bu:Disable();
		bu:Hide();
		return;
	end

	-- Show the 'set' button
	bu:Show();
	val = self:GetNumber();
	if (val < 1) then
		bu:Disable();
	else
		bu:Enable();
	end
end

function EkBomb_OptionsFrame_BufferSizeEB_FocusLost(self)
	-- ----------
	-- Focus lost for the buffer size editbox.
	-- ----------
	self:SetNumber(EkBomb_EventBufferSize);
end

function EkBomb_OptionsFrame_BufferSizeEB_Refresh()
	-- ----------
	-- Update the buffer size editbox/text.
	-- ----------
	if (EkBomb_OptionsFrame_BufferSizeText and EkBomb_OptionsFrame_BufferSizeEB) then
		if (EkBomb_ShowLink == 1) then
			EkBomb_OptionsFrame_BufferSizeText:Show();
			EkBomb_OptionsFrame_BufferSizeEB:Show();
		else
			EkBomb_OptionsFrame_BufferSizeText:Hide();
			EkBomb_OptionsFrame_BufferSizeEB:Hide();
			EkBomb_OptionsFrame_BufferSize_Button:Hide();
		end
	end
end

function EkBomb_OptionsFrame_EnableCB_OnClick(self)
	-- ----------------
	-- User clicked on the 'enable warnings' checkbox.
	-- ----------------
	if (self:GetChecked()) then
		EkBomb_Enable();
	else
		EkBomb_Disable();
	end
end

function EkBomb_OptionsFrame_EnableCB_OnEnter(self)
	-- ----------------
	-- Mouse is over the 'enable warnings' checkbox.
	-- ----------------
	EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_Config_Enable_Tooltip1, EkBomb_TEXT_Config_Enable_Tooltip2);
end

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkCheck
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_EkCheck_CreateTooltipText(index, total, data)
	-- ----------
	-- User moved pointer over a line in the list.
	--
	-- index -- position of the item in the list
	-- total -- number of items in the list
	-- data -- Array of data
	--
	-- 	data[]
	-- 	---------------
	-- 	[1] == EkCheck: Player name
	-- 	[2] == EkCheck: Uppercase english class string (ie. non-localized) (eg. "PRIEST")
	-- 	[3] == EkCheck: 0==Addon not loaded, 1==Addon loaded.
	-- 	[4] == EkCheck: 0==Player has not responded to the check, 1==Player has responded to the check.
	-- 	[5] == EkCheck: Player class (localized)
	-- 	[6] to [10] == EkCheck: Reserved
	--
	-- 	[11] == EkBomb: Version string
	--
	-- Returns:
	--	a) nil -- if no tooltip should be generated.
	--	b) String containing text to appear in the tooltip.
	-- ----------
	-- local text;
	-- local white = "|c00FFFFFF";

	return nil;  -- text;
end

function EkBomb_EkCheck_CreateResponse(data)
	-- ----------------
	-- Create addon specific portion of response string.
	--
	-- data -- Array of data provided by EkCheck.
	--
	-- 	data[]
	-- 	--------
	-- 	[1] == Name of the player that initiated the check.
	-- 	[2] == Name of the addon being checked for.
	--
	-- Returns:
	-- 	a) nil or "" == No addon specific response string.
	--	b) Addon specific response string.
	-- ----------------
	local resp;

	-- EkBomb version
	resp = "v" .. EkBomb_Version;

	-- Return the response string.
	return resp;
end

function EkBomb_EkCheck_ParseResponse(text, data)
	-- ----------------
	-- Parse response text.
	--
	-- text -- The addon specific portion of the response.
	-- data -- Array of data provided by EkCheck.
	--
	-- 	data[]
	-- 	---------------
	-- 	[1] == Player name
	-- 	[2] == Uppercase english class string (ie. non-localized) (eg. "PRIEST")
	-- 	[3] == 0==Addon not loaded, 1==Addon loaded.
	-- 	[4] == 0==Player has not responded to the check, 1==Player has responded to the check.
	-- 	[5] == Player class (localized)
	--
	-- Returns:
	--	a) nil -- Invalid response
	--      b) Array containing addon specific data (to be used for the check window, tooltip, etc).
	-- ----------------
	local strgmatch = string.gmatch or string.gfind;
	local msg;
	local version;

	version = "?";

	msg = strlower(text);

	for k, v in strgmatch(msg, "(%a)(%S+)") do
--		EkBomb_Print("key=[" .. k .. "], val=[" .. v .. "]");

		if (k == "v") then
			version = v;
		end
	end

	return { version };
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* Pre-defined warnings frame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_PredefinedFrame_OnLoad()
	-- ----------
	-- Loading the pre-defined warnings frame.
	-- ----------
	local frame = EkBomb_PredefinedFrame;
	local height, width;

	height = EkBomb_MainFrame:GetHeight();
	width  = EkBomb_MainFrame:GetWidth();

	frame:SetHeight(height - 55);
	frame:SetWidth(width - 50);

	frame.ekTabNum = EkBomb_ListFrame_TabGeneral;
end

function EkBomb_PredefinedFrame_OnShow()
	-- ----------
	-- Showing the pre-defined warnings frame.
	-- ----------
	local frame = EkBomb_PredefinedFrame;

	EkBomb_ListFrame_Configure(frame);
	EkBomb_EditFrame_Configure(frame);

	if (frame.ekTabNum == EkBomb_ListFrame_TabGeneral) then
		EkBomb_ListFrame_Table = EkBomb_PredefinedTable1;
	else
		EkBomb_ListFrame_Table = EkBomb_PredefinedTable1;
	end
	for i = 1, #EkBomb_WarningLists do
		if (EkBomb_WarningLists[i] == EkBomb_ListFrame_Table) then
			EkBomb_ListFrame_TableNum = i;  -- Position of this table in EkBomb_WarningLists
		end
	end

	-- Some defaults in case this is the first time we're showing this list
	EkBomb_ListFrame_NumToDisplay = 16;
	EkBomb_ListFrame_Selected = 0;
	EkBomb_ListFrame_SavedScrollValue = 1;
	EkBomb_ListFrame_MenuNum = 2; -- Menu: View

	-- Restore values from the last time we showed this list
	EkBomb_ListFrame_RestoreValues("PRE" .. frame.ekTabNum);

	EkBomb_ListFrameTab1:Show();
	EkBomb_ListFrameTab1:Hide();  --&&D
	EkBomb_ListFrameTab2:Hide();
	PanelTemplates_SetNumTabs(EkBomb_ListFrame, 1);
	PanelTemplates_SetTab(EkBomb_ListFrame, frame.ekTabNum);

	-- Adjust position of the buttons
	EkBomb_ListFrame_TestButton:ClearAllPoints();
	EkBomb_ListFrame_TestButton:SetPoint("TOPLEFT", "EkBomb_ListFrame", "BOTTOMLEFT", 102, -15);

	EkBomb_ListFrame:Show();
end

function EkBomb_PredefinedFrame_OnHide()
	-- ----------
	-- Hiding the pre-defined warnings frame.
	-- ----------
	EkBomb_ListFrame_SaveValues("PRE" .. EkBomb_PredefinedFrame.ekTabNum);
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* Custom warnings frame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_CustomFrame_OnLoad()
	-- ----------
	-- Loading the custom warnings frame.
	-- ----------
	local frame = EkBomb_CustomFrame;
	local height, width;

	height = EkBomb_MainFrame:GetHeight();
	width  = EkBomb_MainFrame:GetWidth();

	frame:SetHeight(height - 55);
	frame:SetWidth(width - 50);

	frame.ekTabNum = EkBomb_ListFrame_TabGeneral;
end

function EkBomb_CustomFrame_OnShow()
	-- ----------
	-- Showing the custom warnings frame.
	-- ----------
	local frame = EkBomb_CustomFrame;

	EkBomb_ListFrame_Configure(frame);
	EkBomb_EditFrame_Configure(frame);

	if (frame.ekTabNum == EkBomb_ListFrame_TabGeneral) then
		EkBomb_ListFrame_Table = EkBomb_CustomTable1;
	else
		EkBomb_ListFrame_Table = EkBomb_CustomTable2;
	end
	for i = 1, #EkBomb_WarningLists do
		if (EkBomb_WarningLists[i] == EkBomb_ListFrame_Table) then
			EkBomb_ListFrame_TableNum = i;  -- Position of this table in EkBomb_WarningLists
		end
	end

	-- Some defaults in case this is the first time we're showing this list
	EkBomb_ListFrame_NumToDisplay = 16;
	EkBomb_ListFrame_Selected = 0;
	EkBomb_ListFrame_SavedScrollValue = 1;
	EkBomb_ListFrame_MenuNum = 1; -- Menu: Add, Edit, Delete

	-- Restore values from the last time we showed this list
	EkBomb_ListFrame_RestoreValues("CUS" .. frame.ekTabNum);

	EkBomb_ListFrameTab1:Show();
	EkBomb_ListFrameTab2:Show();
	PanelTemplates_SetNumTabs(EkBomb_ListFrame, 2);
	PanelTemplates_SetTab(EkBomb_ListFrame, frame.ekTabNum);

	-- Adjust position of the buttons
	EkBomb_ListFrame_TestButton:ClearAllPoints();
	EkBomb_ListFrame_TestButton:SetPoint("TOPLEFT", "EkBomb_ListFrame", "BOTTOMLEFT", 2, -15);

	EkBomb_ListFrame:Show();
end

function EkBomb_CustomFrame_OnHide()
	-- ----------
	-- Hiding the custom warnings frame.
	-- ----------
	EkBomb_ListFrame_SaveValues("CUS" .. EkBomb_CustomFrame.ekTabNum);
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* List frame (common frame used by both the pre-defined and custom frames).
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_ListFrame_SaveValues(name)
	-- ----------
	-- Save current ListFrame values.
	--
	-- Param: name -- Name to save the values under.
	-- ----------
	local save;

	if (not EkBomb_ListFrame_Saved[name]) then
		EkBomb_ListFrame_Saved[name] = {};
	end
	save = EkBomb_ListFrame_Saved[name];

	save["num"] = EkBomb_ListFrame_NumToDisplay;
	save["sel"] = EkBomb_ListFrame_Selected;
	save["scr"] = EkBomb_ListFrame_ScrollBar:GetValue();
end

function EkBomb_ListFrame_RestoreValues(name)
	-- ----------
	-- Restore saved ListFrame values.
	--
	-- Param: name -- Name the values were saved under.
	-- ----------
	local save;

	if (EkBomb_ListFrame_Saved[name]) then
		save = EkBomb_ListFrame_Saved[name];

		EkBomb_ListFrame_NumToDisplay = save["num"];
		EkBomb_ListFrame_Selected = save["sel"];
		EkBomb_ListFrame_SavedScrollValue = save["scr"];
	end
end

function EkBomb_ListFrame_Configure(frame)
	-- ----------
	-- Configure list frame variables.
	--
	-- Param: frame    -- parent/anchor-to frame.
	-- ----------
	local height = frame:GetHeight();
	local width = frame:GetWidth();

	EkBomb_ListFrame:Hide();
	EkBomb_ListFrame:SetParent(frame);
	EkBomb_ListFrame:ClearAllPoints();
	EkBomb_ListFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, -20);
	EkBomb_ListFrame:SetWidth(width);
	EkBomb_ListFrame:SetHeight(height - 60);
end

function EkBomb_ListFrame_Bombs_Config()
	-- ----------
	-- Verify/Get saved user settings for the predefined list items.
	-- ----------
	local config = EkBomb_BombsConfig;

	for b = 1, #EkBomb_PredefinedTable1 do
		local item = EkBomb_PredefinedTable1[b];
		local key = item["id"];

		-- If user settings table is missing this predefined item...
		if (not config[key]) then
			config[key] = {};
		end

		-- If user settings table is missing the enabled key...
		if (not config[key]["enabled"]) then
			-- Assign default value.
			config[key]["enabled"] = 1;
		end

		-- Update the predefined table with the user setting.
		item["enabled"] = config[key]["enabled"];
	end

	-- Remove old bombs and settings from user config table.
	for k, v in pairs(config) do
		local found = nil;
		for b = 1, #EkBomb_PredefinedTable1 do
			local item = EkBomb_PredefinedTable1[b];
			local key = item["id"];
			if (key == k) then
				found = 1;
				break;
			end
		end
		if (not found) then
			config[k] = nil;
		end
	end
end

function EkBomb_ListFrame_OnLoad(self)
	-- ----------
	-- Loading the list frame.
	-- ----------
	self:SetBackdropBorderColor(0.6, 0.6, 0.6);
	self:SetBackdropColor(0.05, 0.05, 0.05);
end

function EkBomb_ListFrame_OnShow()
	-- ----------
	-- Showing the list frame.
	-- ----------
	EkBomb_ListFrame_ScrollBar.ekheight = EkBomb_ListFrame_NumToDisplay;
	EkBomb_ListFrame_ScrollBar:SetWidth(16);
	EkBomb_ListFrame_ScrollBar:SetHeight(EkBomb_ListFrame_NumToDisplay * EkBomb_ListFrame_Button1:GetHeight() - 29);

	EkBomb_ListFrame_ScrollBarUpdate();
	EkBomb_ListFrame_ScrollBarSetValue(EkBomb_ListFrame_SavedScrollValue);

	EkBomb_ListFrame_ListUpdate();
	EkBomb_ListFrame_ShowMenu();
end

function EkBomb_ListFrame_OnHide()
	-- ----------
	-- Hiding the list frame.
	-- ----------
	EkBomb_ListFrame_SavedScrollValue = EkBomb_ListFrame_ScrollBar:GetValue();
end

function EkBomb_ListFrame_Tab_OnClick(self, tab)
	-- ----------
	-- User clicked one of the tabs at the top of the list frame.
	--
	-- Param: tab -- Tab number (ie. EkBomb_ListFrame_TabGeneral or EkBomb_ListFrame_TabCharacter as
	--               appropriate for the type of list (predfined only supports the general tab)).
	-- ----------
	local frame = EkBomb_ListFrame;
	local pframe = frame:GetParent();

	PanelTemplates_Tab_OnClick(self, frame);

	pframe:Hide();
	pframe.ekTabNum = tab;
	pframe:Show();

	if (tab == 1) then
		EkBomb_ListFrameTab1:ClearAllPoints();
		EkBomb_ListFrameTab1:SetPoint("BOTTOMLEFT", "EkBomb_ListFrame", "TOPLEFT", 10, -3);
		EkBomb_ListFrameTab2:ClearAllPoints();
		EkBomb_ListFrameTab2:SetPoint("LEFT", "EkBomb_ListFrameTab1", "RIGHT", 0, 1);
	else
		EkBomb_ListFrameTab1:ClearAllPoints();
		EkBomb_ListFrameTab1:SetPoint("BOTTOMLEFT", "EkBomb_ListFrame", "TOPLEFT", 10, -1);
		EkBomb_ListFrameTab2:ClearAllPoints();
		EkBomb_ListFrameTab2:SetPoint("LEFT", "EkBomb_ListFrameTab1", "RIGHT", 0, -3);
	end
end

function EkBomb_ListFrame_Sort(list, sel)
	-- ----------
	-- Sort the list.
	--
	-- Param: list -- (optional) Table to sort (default is EkBomb_ListFrame_Table).
	-- Param: sel -- (optional) Subscript of currently selected item in the list.
	-- Returns: Subscript of the selected item after the sort, or nil.
	-- ----------
	if (not list) then
		list = EkBomb_ListFrame_Table;
	end

	local num = #list;

	if (sel and sel ~= 0) then
		-- Mark the selected item so we can find it after the sort.
		list[sel].mark = 1;
	end

	sort(list,
		function (a, b)
			return strlower(a.name) < strlower(b.name);
		end
	);

	if (sel and sel ~= 0) then
		-- Find the marked item (ie. the selected item).
		for i = 1, num do
			if (list[i].mark) then
				list[i].mark = nil;
				return i;
			end
		end
	end

	return nil;
end

function EkBomb_ListFrame_ScrollBarUpdate()
	-- ----------
	-- Update the scroll bar.
	-- ----------
	local scrollBar = EkBomb_ListFrame_ScrollBar;
	local list = EkBomb_ListFrame_Table;
	local NumItems = #list;
	local showScrollBar = nil;
	local ItemOffset;
	local maxVal;

	if ( NumItems > EkBomb_ListFrame_NumToDisplay ) then
		showScrollBar = 1;
		maxVal = NumItems - EkBomb_ListFrame_NumToDisplay + 1;
		scrollBar:SetMinMaxValues(1, maxVal)
		scrollBar:Show();
		ItemOffset = scrollBar:GetValue() - 1;
	else
		maxVal = 1;
		scrollBar:SetMinMaxValues(1, maxVal);
		scrollBar:Hide();
		ItemOffset = 0;
	end

	scrollBar.ekoffset = ItemOffset;
	return ItemOffset, showScrollBar;
end

function EkBomb_ListFrame_ScrollBarSetValue(val)
	-- ----------
	-- Set the scrollbar value.
	-- ----------
	local scrollBar = EkBomb_ListFrame_ScrollBar;
	local minVal, maxVal = scrollBar:GetMinMaxValues();
	if (val > maxVal) then
		val = maxVal;
	elseif (val < 1) then
		val = 1;
	end
	scrollBar:SetValue(val);
end

function EkBomb_ListFrame_ScrollBarSetMiddle(sub)
	-- ----------
	-- Set the scrollbar so that the specified subscript is
	-- as close to the middle of the window as possible.
	--
	-- Param: sub -- (Number) Subscript of the item that should appear
	--               as close to the middle of the window as possible.
	-- ----------
	local scrollBar = EkBomb_ListFrame_ScrollBar;
	local minVal, maxVal = scrollBar:GetMinMaxValues();
	local mid = floor(EkBomb_ListFrame_NumToDisplay / 2);
	local val = sub - mid + 1;
	EkBomb_ListFrame_ScrollBarSetValue(val);
end

function EkBomb_ListFrame_ScrollBarSetBottom(sub)
	-- ----------
	-- Set the scrollbar so that the specified subscript is
	-- as close to the bottom of the window as possible.
	--
	-- Param: sub -- (Number) Subscript of the item.
	-- ----------
	local val = sub - EkBomb_ListFrame_NumToDisplay + 1;
	if (val < 1) then
		val = 1;
	end
	EkBomb_ListFrame_ScrollBarSetValue(val);
end

function EkBomb_ListFrame_ListUpdate()
	-- ----------
	-- Update the displayed data.
	-- ----------
	local width;
	local list, NumItems, sel;
	local ItemIndex, ItemOffset;
	local scrollBar, showScrollBar;
	local button, rec, name, high, cb;

	if (not EkBomb_ListFrame:IsShown()) then
		return;
	end

	width = EkBomb_ListFrame:GetWidth();

	list = EkBomb_ListFrame_Table;
	NumItems = #list;

	sel = EkBomb_ListFrame_Selected;

	scrollBar = EkBomb_ListFrame_ScrollBar;
	showScrollBar = nil;

	ItemOffset, showScrollBar = EkBomb_ListFrame_ScrollBarUpdate();

	EkBomb_ListFrame.ekTopSub = 0;
	EkBomb_ListFrame.ekLastSub = 0;

	for i = 1, EkBomb_ListFrame_NumToDisplay do
		ItemIndex = ItemOffset + i;
		button = _G["EkBomb_ListFrame_Button" .. i];
		cb = _G["EkBomb_ListFrame_Button" .. i .. "CB"];
		name = _G["EkBomb_ListFrame_Button" .. i .. "Name"];
		high = _G["EkBomb_ListFrame_Button" .. i .. "Highlight"];
		button.ItemIndex = ItemIndex;

		if ( showScrollBar ) then
			button:SetWidth(width - 40);
			EkBomb_ListFrame_Tooltip_OffsetRight = 25;
		else
			button:SetWidth(width - 20);
			EkBomb_ListFrame_Tooltip_OffsetRight = 5;
		end
		name:SetWidth(button:GetWidth() - 30);

		if (ItemIndex > 0 and ItemIndex <= NumItems) then
			rec = list[ItemIndex];

			cb:SetChecked(rec["enabled"]);
			cb:Show();

			name:SetText(rec["name"]);
			name:Show();

			-- Highlight the correct name
			if ( ItemIndex == sel ) then
				high:Show();
			else
				high:Hide();
			end

			button:Show();

			if (EkBomb_ListFrame.ekTopSub == 0) then
				EkBomb_ListFrame.ekTopSub = ItemIndex;
			end
			EkBomb_ListFrame.ekLastSub = ItemIndex;
		else
			button:Hide();
		end
	end

	scrollBar:SetValue(scrollBar:GetValue());
end

function EkBomb_ListFrame_IsItemVisible(sub)
	-- ----------
	-- Returns true if specified subscript is visible in the list window.
	-- ----------
	if (not sub or sub == 0 or EkBomb_ListFrame.ekTopSub == 0 or EkBomb_ListFrame.ekLastSub == 0) then
		return false;
	end
	if (sub >= EkBomb_ListFrame.ekTopSub and sub <= EkBomb_ListFrame.ekLastSub) then
		return true;
	end
	return false;
end

function EkBomb_ListFrame_Button_OnClick(self, button)
	-- ----------
	-- User clicked on an item in the list.
	--
	-- Param: button == Which mouse button was pushed.
	-- ----------
	local sub = self.ItemIndex; -- Subscript stored in this button

	if (button == "LeftButton") then
		-- Select a single item. Cancels any existing selection.
		EkBomb_ListFrame_Selected = sub;
		EkBomb_ListFrame_ListUpdate();
		EkBomb_ListFrame_ShowMenu();

	elseif (button == "RightButton") then
		-- Toggle the current item's enabled status.
		local list = EkBomb_ListFrame_Table;
		local item = list[sub];
		local cb = _G["EkBomb_ListFrame_Button" .. sub .. "CB"];

		if (item["enabled"] == 1) then
			EkBomb_ListFrame_SetEnabled(sub, 0);
			cb:SetChecked(0);
		else
			EkBomb_ListFrame_SetEnabled(sub, 1);
			cb:SetChecked(1);
		end

		EkBomb_ListFrame_ListUpdate();
	end
end

function EkBomb_ListFrame_Button_OnDoubleClick(self, button)
	-- ----------
	-- User double clicked on an item in the list.
	-- 
	-- Param: button == Which mouse button was pushed.
	-- ----------
	local sub = self.ItemIndex; -- Subscript stored in this button

	if (button == "LeftButton") then
		-- Select and edit the item.
		EkBomb_ListFrame_Selected = sub;
		EkBomb_ListFrame_ListUpdate();
		EkBomb_ListFrame_ShowMenu();

		local frame = EkBomb_ListFrame:GetParent();
		local edit;
		if (frame:GetName() == "EkBomb_PredefinedFrame") then
			edit = 0;
		else
			edit = 1;
		end
		EkBomb_ListFrame_EditItem(edit);
	end
end

function EkBomb_ListFrame_Button_OnEnter(self)
	-- ----------
	-- User moved pointer over an item.
	-- ----------
	local sub = self.ItemIndex; -- Subscript stored in this button
	local warnType;

	local xp = "LEFT";
	local yp = "BOTTOM";
	local xo = -6;
	local xthis, ythis = self:GetCenter();
	local xui, yui = UIParent:GetCenter();
	if (xthis < xui) then
		xp = "RIGHT";
		xo = EkBomb_ListFrame_Tooltip_OffsetRight;
	end
	if (ythis < yui) then
--		yp = "TOP";
		yp = "";
		if (xthis < xui) then
			xp = "RIGHT";
		else
			xp = "LEFT";
		end
	end

	local item;
	item = EkBomb_ListFrame_Table[sub];

	local text, white;

	white = "|c00FFFFFF";
	text = "";

        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Name .. "|r " .. item.name .. "\n";
        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Small .. "|r " .. item.small .. "\n";
        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Large .. "|r " .. item.large .. "\n";
        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Duration .. "|r " .. item.timer .. "\n";

        text = text .. "\n";

        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Click1 .. "|r " .. EkBomb_TEXT_ListFrame_Tooltip_Click2 .. "\n";

--!!	text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Double_Click1 .. "|r ";

	local frame = EkBomb_ListFrame:GetParent();
	if (frame:GetName() == "EkBomb_PredefinedFrame") then
--!!	        text = text .. EkBomb_TEXT_ListFrame_Tooltip_Double_Click3 .. "\n";
		warnType = EkBomb_TEXT_ListFrame_Tooltip_Predefined;
	else
--!!	        text = text .. EkBomb_TEXT_ListFrame_Tooltip_Double_Click2 .. "\n";
		warnType = EkBomb_TEXT_ListFrame_Tooltip_Custom;
	end

        text = text .. white .. EkBomb_TEXT_ListFrame_Tooltip_Right_Click1 .. "|r " .. EkBomb_TEXT_ListFrame_Tooltip_Right_Click2 .. "\n";

--	EkBomb_SetTooltip_Anchor(self, text, "ANCHOR_" .. yp .. xp, xo, 0, 1);

	EkBomb_SetTooltip_Corner(self, warnType, text);
end

function EkBomb_ListFrame_CB_OnClick(cb)
	-- ----------
	-- User clicked on a checkbox to the left of the item name.
	--
	-- Param: cb == Checkbox object
	-- ----------
	local button = cb:GetParent();
	local sub = button.ItemIndex; -- Subscript stored in this button

	if (cb:GetChecked()) then
		EkBomb_ListFrame_SetEnabled(sub, 1);
	else
		EkBomb_ListFrame_SetEnabled(sub, 0);
	end
end

function EkBomb_ListFrame_SetEnabled(sub, enabled)
	-- ----------
	-- Set the enabled flag to 0 or 1 for the specified subscript.
	--
	-- Param: sub == Subscript number of the list
	-- Param: enabled == 0 (disable) or 1 (enable)
	-- ----------
	local list = EkBomb_ListFrame_Table;
	local item = list[sub];
	local old;

	old = item["enabled"];
	item["enabled"] = enabled;

	if (old ~= enabled) then
		EkBomb_EventList_UpdateWarningList(EkBomb_ListFrame_TableNum);
	end

	-- If the predefined table...
	if (item["id"]) then
		-- Update user setting table.
		local config = EkBomb_BombsConfig;
		local key = item["id"];

		if (not config[key]) then
			config[key] = {};
		end
		config[key]["enabled"] = enabled;
	end
end

function EkBomb_ListFrame_ShowMenu()
	-- ----------
	-- Show the list frame button menu.
	-- ----------
	local list = EkBomb_ListFrame_Table;
	local sel = EkBomb_ListFrame_Selected;

	if (EkBomb_ListFrame_MenuNum == 1) then
		-- Custom frame: Add, Edit, Delete
		EkBomb_ListFrame_TestButton:Show();
		EkBomb_ListFrame_ViewButton:Hide();
		EkBomb_ListFrame_AddButton:Show();
		EkBomb_ListFrame_EditButton:Show();
		EkBomb_ListFrame_DeleteButton:Show();
		EkBomb_ListFrame_YesNoMsg:Hide();
		EkBomb_ListFrame_YesButton:Hide();
		EkBomb_ListFrame_NoButton:Hide();
		if (sel == 0) then
			EkBomb_ListFrame_EditButton:Disable();
			EkBomb_ListFrame_DeleteButton:Disable();
		else
			EkBomb_ListFrame_EditButton:Enable();
			EkBomb_ListFrame_DeleteButton:Enable();
		end
		if (#list == 0) then
			EkBomb_ListFrame_TestButton:Disable();
		else
			EkBomb_ListFrame_TestButton:Enable();
		end

	elseif (EkBomb_ListFrame_MenuNum == 2) then
		-- Predefined frame: View
		EkBomb_ListFrame_TestButton:Show();
--!!		EkBomb_ListFrame_ViewButton:Show();
		EkBomb_ListFrame_AddButton:Hide();
		EkBomb_ListFrame_EditButton:Hide();
		EkBomb_ListFrame_DeleteButton:Hide();
		EkBomb_ListFrame_YesNoMsg:Hide();
		EkBomb_ListFrame_YesButton:Hide();
		EkBomb_ListFrame_NoButton:Hide();
		if (sel == 0) then
			EkBomb_ListFrame_ViewButton:Disable();
		else
			EkBomb_ListFrame_ViewButton:Enable();
		end
		if (#list == 0) then
			EkBomb_ListFrame_TestButton:Disable();
		else
			EkBomb_ListFrame_TestButton:Enable();
		end

	elseif (EkBomb_ListFrame_MenuNum == 3) then
		-- Delete item?
		EkBomb_ListFrame_TestButton:Hide();
		EkBomb_ListFrame_ViewButton:Hide();
		EkBomb_ListFrame_AddButton:Hide();
		EkBomb_ListFrame_EditButton:Hide();
		EkBomb_ListFrame_DeleteButton:Hide();
		EkBomb_ListFrame_YesNoMsg:SetText(EkBomb_TEXT_ListFrame_ConfirmDelete);
		EkBomb_ListFrame_YesNoMsg:Show();
		EkBomb_ListFrame_YesButton:Show();
		EkBomb_ListFrame_NoButton:Show();
	end
end

function EkBomb_ListFrame_TestButton_OnClick()
	-- ----------
	-- User clicked on the Test button.
	-- ----------
	local rand, sub;
	local list = EkBomb_ListFrame_Table;
	local sel  = EkBomb_ListFrame_Selected;

	if (#list == 0) then
		return;
	end

	-- Allow user to click test button when nothing is selected (ie. sel == 0)

	EkBomb_Hide();

	if (IsShiftKeyDown()) then
		-- Sequential item
		rand = nil;
		sub  = nil;
		EkBomb_TestIndex = sel;  -- Start at item after selected item.

	elseif (IsControlKeyDown()) then
		-- Random item
		rand = 1;
		sub  = nil;
	else
		-- Specific item
		rand = nil;
		sub = sel;
		if (sub == 0) then
			sub = 1;
		end
	end

	EkBomb_TestWarning(rand, EkBomb_ListFrame_TableNum, sub, nil, 1);

	-- After showing the warning, change selection to the warning
	-- that was shown and update the list window
	EkBomb_ListFrame_Selected = EkBomb_TestIndex;

	if (not sub or not EkBomb_ListFrame_IsItemVisible(sub)) then
		-- Showed random or sequential warning, or shown item is not currently visible.
		EkBomb_ListFrame_ScrollBarSetMiddle(EkBomb_ListFrame_Selected);
	end

	EkBomb_ListFrame_ListUpdate();
	EkBomb_ListFrame_ShowMenu();
end

function EkBomb_ListFrame_ViewButton_OnClick()
	-- ----------
	-- User clicked on the View button.
	-- ----------
	EkBomb_ListFrame_EditItem(0);
end

function EkBomb_ListFrame_AddButton_OnClick()
	-- ----------
	-- User clicked on the Add button.
	-- ----------
	local item;

	EkBomb_ListFrame:Hide();

	-- Initialize the edit fields
	item = {};
	EkBomb_EditFrame_NewSettings(item);
	EkBomb_EditFrame_PutSettings(item);

	EkBomb_EditFrame_EditMode = 2;  -- 0==View, 1==Edit, 2==Add

	EkBomb_EditFrame:Show();
	EkBomb_EditFrame_NameEB:SetFocus();
end

function EkBomb_ListFrame_EditButton_OnClick()
	-- ----------
	-- User clicked on the Edit button.
	-- ----------
	EkBomb_ListFrame_EditItem(1);
end

function EkBomb_ListFrame_DeleteButton_OnClick()
	-- ----------
	-- User clicked on the Delete button.
	-- ----------
	EkBomb_ListFrame_MenuNum = 3;  -- Menu: Delete item?  Yes  No
	EkBomb_ListFrame_ShowMenu();
end

function EkBomb_ListFrame_YesButton_OnClick()
	-- ----------
	-- User clicked on the Yes button.
	-- ----------
	local list = EkBomb_ListFrame_Table;
	local sel  = EkBomb_ListFrame_Selected;
	if (sel == 0) then
		return;
	end

	-- Delete the item from the list.
	tremove(list, sel);

	-- Adjust the selection.
	if (sel > #list) then
		sel = #list;
	end
	EkBomb_ListFrame_Selected = sel;

	-- Rebuild the event list entries for this table.
	EkBomb_EventList_UpdateWarningList(EkBomb_ListFrame_TableNum);

	EkBomb_ListFrame_ScrollBarUpdate();
	EkBomb_ListFrame_ScrollBarSetValue(EkBomb_ListFrame_Selected);
	EkBomb_ListFrame_ListUpdate();

	EkBomb_ListFrame_MenuNum = 1;  -- Custom frame menu: Add, Edit, Delete
	EkBomb_ListFrame_ShowMenu();
end

function EkBomb_ListFrame_NoButton_OnClick()
	-- ----------
	-- User clicked on the No button.
	-- ----------
	EkBomb_ListFrame_MenuNum = 1;  -- Custom frame menu: Add, Edit, Delete
	EkBomb_ListFrame_ShowMenu();
end

function EkBomb_ListFrame_EditItem(edit)
	-- ----------
	-- Edit the selected item.
	--
	-- Param: edit == 0 (view), 1 (edit)
	-- ----------
	local list = EkBomb_ListFrame_Table;
	local sel  = EkBomb_ListFrame_Selected;
	if (sel == 0) then
		return;
	end
	local item = list[sel];

	EkBomb_ListFrame:Hide();

	-- Initialize the edit fields
	EkBomb_EditFrame_PutSettings(item);

	EkBomb_EditFrame_EditMode = edit;  -- 0==View, 1==Edit, 2==Add

	EkBomb_EditFrame:Show();
	if (EkBomb_EditFrame_EditMode == 1) then
		EkBomb_EditFrame_NameEB:SetFocus();
	end
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* Edit frame
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_EditFrame_Configure(frame)
	-- ----------
	-- Configure edit frame variables.
	--
	-- Param: frame -- Parent frame.
	-- ----------
	local height;

	height = frame:GetHeight();

	EkBomb_EditFrame:Hide();
	EkBomb_EditFrame:SetParent(frame);
	EkBomb_EditFrame:ClearAllPoints();
	EkBomb_EditFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, -5);
	EkBomb_EditFrame:SetWidth(frame:GetWidth());
	EkBomb_EditFrame:SetHeight(height - 45);

	EkBomb_EditFrame_NameEB:ClearAllPoints();
	EkBomb_EditFrame_NameEB:SetPoint("TOPLEFT", "EkBomb_EditFrame", "TOPLEFT", EkBomb_EditFrame_Offset, -20);

	EkBomb_EditFrame_EB_MaxWidth = EkBomb_EditFrame:GetWidth() - EkBomb_EditFrame_Offset - 30;
end

function EkBomb_EditFrame_OnLoad(self)
	-- ----------
	-- Loading the edit frame.
	-- ----------
	self:SetBackdropBorderColor(0.6, 0.6, 0.6);
	self:SetBackdropColor(0.10, 0.10, 0.10);
end

function EkBomb_EditFrame_OnShow()
	-- ----------
	-- Showing the edit frame.
	-- ----------
	EkBomb_EditFrame_ArgEB_Validate(EkBomb_EditFrame_ArgEB, nil);
	EkBomb_EditFrame_ShowHideItems();
	EkBomb_EditFrame_ShowMenu();

	local frame = EkBomb_EditFrame:GetParent();
	local msg;

	if (EkBomb_EditFrame_EditMode == 0) then
		msg = EkBomb_TEXT_ListFrame_View_Button;
	elseif (EkBomb_EditFrame_EditMode == 1) then
		msg = EkBomb_TEXT_ListFrame_Edit_Button;
	else
		msg = EkBomb_TEXT_ListFrame_Add_Button;
	end
	msg = msg .. " (";
	if (frame.ekTabNum == EkBomb_ListFrame_TabGeneral) then
		msg = msg .. EkBomb_TEXT_ListFrame_GeneralButton;
	else
		msg = msg .. string.format(EkBomb_TEXT_ListFrame_CharacterButton, (UnitName("player")));
	end
	msg = msg .. ")";
	EkBomb_EditFrame_Subtitle:SetText(msg);
end

function EkBomb_EditFrame_ShowMenu()
	-- ----------
	-- Show the edit frame button menu.
	-- ----------
	EkBomb_EditFrame_CopyButton:Show();
	EkBomb_EditFrame_PasteButton:Show();
	EkBomb_EditFrame_OkButton:Show();
	EkBomb_EditFrame_CancelButton:Show();

	if (not EkBomb_EditFrame_Clipboard) then
		EkBomb_EditFrame_PasteButton:Disable();
	else
		EkBomb_EditFrame_PasteButton:Enable();
	end

	if (not EkBomb_EditFrame_IsValid()) then
		EkBomb_EditFrame_OkButton:Disable();
	else
		EkBomb_EditFrame_OkButton:Enable();
	end

	EkBomb_EditFrame_CancelButton:Enable();

	if (EkBomb_EditFrame_EditMode == 0) then
		-- Viewing
		EkBomb_EditFrame_CopyButton:ClearAllPoints();
		EkBomb_EditFrame_CopyButton:SetPoint("TOPLEFT", "EkBomb_EditFrame", "BOTTOMLEFT", 102, -15);
		EkBomb_EditFrame_PasteButton:Hide();
		EkBomb_EditFrame_OkButton:Hide();
		EkBomb_EditFrame_CancelButton:ClearAllPoints();
		EkBomb_EditFrame_CancelButton:SetPoint("LEFT", "EkBomb_EditFrame_CopyButton", "RIGHT", 5, 0);
	else
		-- Editing, Adding
		EkBomb_EditFrame_CopyButton:ClearAllPoints();
		EkBomb_EditFrame_CopyButton:SetPoint("TOPLEFT", "EkBomb_EditFrame", "BOTTOMLEFT", 2, -15);
		EkBomb_EditFrame_CancelButton:ClearAllPoints();
		EkBomb_EditFrame_CancelButton:SetPoint("LEFT", "EkBomb_EditFrame_OkButton", "RIGHT", 5, 0);
	end
end

function EkBomb_EditFrame_IsValid()
	-- ----------
	-- Validate edit frame contents.
	-- ----------
	local name, event;

	-- Warning name must not be blank.
	name = EkBomb_EditFrame_NameEB:GetText();
	if (string.find(name, "^%s*$")) then
		return false;
	end

	-- Event name must not be blank.
	event = EkBomb_EditFrame_EventEB:GetText();
	if (string.find(event, "^%s*$")) then
		return false;
	end

	return true;
end

function EkBomb_EditFrame_NewSettings(item)
	-- ----------
	-- Initialize table of warning settings to default values.
	--
	-- Param: item -- Table to hold settings.
	-- ----------
	item["name"]   = EkBomb_TEXT_EditFrame_DefaultWarningName;
	item["small"]  = EkBomb_TEXT_EditFrame_DefaultSmallText;
	item["large"]  = EkBomb_TEXT_EditFrame_DefaultLargeText;
	item["timer"]  = 0;
	item["event"]  = "";
	item["arg"]    = 1;
	item["search"] = EkBomb_SearchCode_Exact;
	item["value"]  = "";
	item["class"]  = "ANY";
	item["zone"]   = GetRealZoneText();

	item["enabled"] = 0;
end

function EkBomb_EditFrame_GetSettings(item)
	-- ----------
	-- Copy currently displayed warning settings into specified table.
	--
	-- Param: item -- Table to hold settings.
	-- ----------
	local value;

	item["name"]   = EkBomb_EditFrame_NameEB:GetText();
	item["small"]  = EkBomb_EditFrame_SmallEB:GetText();
	item["large"]  = EkBomb_EditFrame_LargeEB:GetText();
	item["timer"]  = EkBomb_EditFrame_TimerEB:GetNumber();
	item["event"]  = strupper(EkBomb_EditFrame_EventEB:GetText());
	item["arg"]    = EkBomb_EditFrame_ArgEB:GetNumber();
	item["search"] = EkBomb_EditFrame_GetSearchCode();
	item["class"]  = EkBomb_EditFrame_GetClassCode();
	item["zone"]   = EkBomb_EditFrame_ZoneEB:GetText();

	-- UnConvert certain characters in the edited search value.
	value = EkBomb_EditFrame_SearchValueEB:GetText();
	value = string.gsub(value, "||",  "|");
	value = string.gsub(value, "\\n", "\n");
	item["value"]  = value;
end

function EkBomb_EditFrame_PutSettings(item)
	-- ----------
	-- Copy settings from the specified table into the currently displayed warning.
	--
	-- Param: item -- Table containing settings.
	-- ----------
	local value;

	-- Convert certain characters in the search value so that they display and edit properly.
	value = item["value"];
	value = string.gsub(value, "|",  "||");
	value = string.gsub(value, "\n", "\\n");
	EkBomb_EditFrame_SearchValueEB:SetText(value);

	EkBomb_EditFrame_NameEB:SetText(item["name"]);
	EkBomb_EditFrame_SmallEB:SetText(item["small"]);
	EkBomb_EditFrame_LargeEB:SetText(item["large"]);
	EkBomb_EditFrame_TimerEB:SetNumber(item["timer"]);
	EkBomb_EditFrame_EventEB:SetText(item["event"]);
	EkBomb_EditFrame_ArgEB:SetNumber(item["arg"]);
	EkBomb_EditFrame_SetSearchCode(item["search"]);
	EkBomb_EditFrame_SetClassCode(item["class"]);
	EkBomb_EditFrame_ZoneEB:SetText(item["zone"]);
end

-- ----------
-- EditFrame edit boxes
-- ----------

function EkBomb_EditFrame_EB_Get_OnTab(self)
	-- ----------
	-- When tab is pressed, get previous/next edit box.
	--
	-- Returns: reference to editbox to use, or nil.
	-- ---------
	local eb, found, i;
	local name = self:GetName();

	-- Locate the edit box
	found = nil;
	i = 1;
	repeat
		if (name == EkBomb_EditFrame_EB_List[i]) then
			found = i;
			break;
		end
		i = i + 1;
	until (i > #EkBomb_EditFrame_EB_List);

	if (found) then
		i = found;
		if (IsShiftKeyDown()) then
			repeat
				i = i - 1;
				if (i < 1) then
					i = #EkBomb_EditFrame_EB_List;
				end
				eb = _G[ (EkBomb_EditFrame_EB_List[i]) ];
			until (eb and eb:IsShown());
		else
			repeat
				i = i + 1;
				if (i > #EkBomb_EditFrame_EB_List) then
					i = 1;
				end
				eb = _G[ (EkBomb_EditFrame_EB_List[i]) ];
			until (eb and eb:IsShown());
		end
	end

	return eb;
end

function EkBomb_EditFrame_TimerEB_Validate(eb, abort)
	-- ----------
	-- Validate contents of the timer editbox.
	-- ----------
	local val;
	val = eb:GetNumber();
	if (val < 0) then
		val = 0;
	end
	eb:SetNumber(val);
end

function EkBomb_EditFrame_EventEB_Validate(eb, abort)
	-- ----------
	-- Validate contents of the event editbox.
	-- ----------
	eb:SetText(strupper(eb:GetText()));
end

function EkBomb_EditFrame_ArgEB_Validate(eb, abort)
	-- ----------
	-- Validate contents of the arg editbox.
	-- ----------
	local val;
	val = eb:GetNumber();
	if (val < 0) then
		val = 0;
	elseif (val > 9) then
		val = 9;
	end
	eb:SetNumber(val);
	EkBomb_EditFrame_ArgEB_TextChanged(eb);
end

function EkBomb_EditFrame_ArgEB_TextChanged(eb)
	-- ----------
	-- Contents of the arg editbox have changed.
	-- ----------
	local val;
	val = eb:GetNumber();
	EkBomb_EditFrame_ShowHideItems();
end

function EkBomb_EditFrame_ShowHideItems()
	-- ----------
	-- Show/hide certain items on the edit frame depening on values chosen by user.
	-- ----------
	if (EkBomb_EditFrame_ArgEB:GetNumber() == 0) then
		-- Use not searching for an argument, so hide the search drop down,
		-- and the search edit box.
		EkBomb_EditFrame_SearchDropDown:Hide();
		EkBomb_EditFrame_SearchValueEB:Hide();

		EkBomb_EditFrame_ClassDropDown:ClearAllPoints();
		EkBomb_EditFrame_ClassDropDown:SetPoint("TOPLEFT", "EkBomb_EditFrame_ArgEB", "BOTTOMLEFT", -24, -10);
	else
		-- User is searching for an argument, so show the search drop down,
		-- and the search edit box.
		EkBomb_EditFrame_SearchDropDown:Show();
		EkBomb_EditFrame_SearchValueEB:Show();

		EkBomb_EditFrame_ClassDropDown:ClearAllPoints();
		EkBomb_EditFrame_ClassDropDown:SetPoint("TOPLEFT", "EkBomb_EditFrame_SearchValueEB", "BOTTOMLEFT", -24, -15);
	end
end

-- ----------
-- EditFrame Search Type drop down menu
-- ----------

function EkBomb_EditFrame_SearchDropDown_OnLoad()
	-- ----------
	-- Load the dropdown menu.
	-- ----------
	local id = 1;
	local dd = EkBomb_EditFrame_SearchDropDown;

	UIDropDownMenu_Initialize(dd, EkBomb_EditFrame_SearchDropDown_Initialize);
	UIDropDownMenu_SetWidth(dd, 150);

	EkBomb_EditFrame_SearchDropDown_SetID(id);
end

function EkBomb_EditFrame_SearchDropDown_Initialize()
	-- ----------
	-- Initialize the dropdown menu.
	-- ----------
	local list = EkBomb_EditFrame_SearchType_Table;
	local info;
	for key, val in pairs(list) do
		info = {};
		info.text = val.name;
		info.func = EkBomb_EditFrame_SearchDropDown_OnClick;
		UIDropDownMenu_AddButton(info);
	end
end

function EkBomb_EditFrame_SearchDropDown_OnClick(self)
	-- ----------
	-- User has selected an item from the dropdown menu.
	-- ----------
	EkBomb_EditFrame_SearchDropDown_SetID(self:GetID());  -- self == drop down button
	EkBomb_EditFrame_ShowHideItems();
end

function EkBomb_EditFrame_SearchDropDown_SetID(id)
	-- ----------
	-- Set the current value of the dropdown menu.
	-- ----------
	local list = EkBomb_EditFrame_SearchType_Table;
	local dd = EkBomb_EditFrame_SearchDropDown;

	UIDropDownMenu_SetSelectedID(dd, id);
	UIDropDownMenu_SetText(dd, list[id].name);

	local ebt = EkBomb_EditFrame_SearchValueEB_LeftText;

	if (ebt) then
		if (list[id].code == EkBomb_SearchCode_Exact) then
			ebt:SetText(EkBomb_TEXT_EditFrame_ExactValue_Text);
			EkBomb_TEXT_EditFrame_SearchValue_Tooltip1 = EkBomb_TEXT_EditFrame_ExactValue_Tooltip1;
			EkBomb_TEXT_EditFrame_SearchValue_Tooltip2 = EkBomb_TEXT_EditFrame_ExactValue_Tooltip2;
		else
			ebt:SetText(EkBomb_TEXT_EditFrame_PatternValue_Text);
			EkBomb_TEXT_EditFrame_SearchValue_Tooltip1 = EkBomb_TEXT_EditFrame_PatternValue_Tooltip1;
			EkBomb_TEXT_EditFrame_SearchValue_Tooltip2 = EkBomb_TEXT_EditFrame_PatternValue_Tooltip2;
		end
	end
end

function EkBomb_EditFrame_SearchDropDown_GetID()
	-- ----------
	-- Gett the current value of the dropdown menu.
	-- ----------
	local dd = EkBomb_EditFrame_SearchDropDown;
	return UIDropDownMenu_GetSelectedID(dd);
end

function EkBomb_EditFrame_GetSearchCode()
	-- ----------
	-- Get search type code.
	-- ----------
	local id;
	id = EkBomb_EditFrame_SearchDropDown_GetID();
	if (not id) then
		id = 1;
	end
	return EkBomb_EditFrame_SearchType_Table[id].code;
end

function EkBomb_EditFrame_SetSearchCode(code)
	-- ----------
	-- Set search type using the specified search type code.
	-- ----------
	local id;
	id = EkBomb_EditFrame_LookupSearchTypeCode(code);
	if (not id) then
		id = 1;
	end
	EkBomb_EditFrame_SearchDropDown_SetID(id);
end

function EkBomb_EditFrame_LookupSearchTypeCode(code)
	-- ----------
	-- Look up code and return table subscript.
	--
	-- Param: code
	-- Returns: table subscript
	-- ----------
	local list = EkBomb_EditFrame_SearchType_Table;
	for i = 1, #list do
		if (list[i].code == code) then
			return i;
		end
	end
	return nil;
end

-- ----------
-- EditFrame Class drop down menu
-- ----------

function EkBomb_EditFrame_ClassDropDown_OnLoad()
	-- ----------
	-- Load the dropdown menu.
	-- ----------
	local dd = EkBomb_EditFrame_ClassDropDown;

	UIDropDownMenu_Initialize(dd, EkBomb_EditFrame_ClassDropDown_Initialize);
	UIDropDownMenu_SetWidth(dd, 150);

	EkBomb_EditFrame_ClassDropDown_SetID(1);
end

function EkBomb_EditFrame_ClassDropDown_Initialize()
	-- ----------
	-- Initialize the dropdown menu.
	-- ----------
	local list = EkBomb_EditFrame_Class_Table;
	local info;
	for key, val in pairs(list) do
		info = {};
		info.text = val.name;
		info.func = EkBomb_EditFrame_ClassDropDown_OnClick;
		UIDropDownMenu_AddButton(info);
	end
end

function EkBomb_EditFrame_ClassDropDown_OnClick(self)
	-- ----------
	-- User has selected an item from the dropdown menu.
	-- ----------
	EkBomb_EditFrame_ClassDropDown_SetID(self:GetID());  -- self == drop down button
end

function EkBomb_EditFrame_ClassDropDown_SetID(id)
	-- ----------
	-- Set the current value of the dropdown menu.
	-- ----------
	local list = EkBomb_EditFrame_Class_Table;
	local dd = EkBomb_EditFrame_ClassDropDown;
	UIDropDownMenu_SetSelectedID(dd, id);
	UIDropDownMenu_SetText(dd, list[id].name);
end

function EkBomb_EditFrame_ClassDropDown_GetID()
	-- ----------
	-- Gett the current value of the dropdown menu.
	-- ----------
	local dd = EkBomb_EditFrame_ClassDropDown;
	return UIDropDownMenu_GetSelectedID(dd);
end

function EkBomb_EditFrame_GetClassCode()
	-- ----------
	-- Get class code.
	-- ----------
	local id;
	id = EkBomb_EditFrame_ClassDropDown_GetID();
	if (not id) then
		id = 1;
	end
	return EkBomb_EditFrame_Class_Table[id].code;
end

function EkBomb_EditFrame_SetClassCode(code)
	-- ----------
	-- Set class using the specified class code.
	-- ----------
	local id;
	id = EkBomb_EditFrame_LookupClassCode(code);
	if (not id) then
		id = 1;
	end
	EkBomb_EditFrame_ClassDropDown_SetID(id);
end

function EkBomb_EditFrame_LookupClassCode(code)
	-- ----------
	-- Look up code and return table subscript.
	--
	-- Param: code (eg. "ANY", "DRUID")
	-- Returns: table subscript
	-- ----------
	local list = EkBomb_EditFrame_Class_Table;
	for i = 1, #list do
		if (list[i].code == code) then
			return i;
		end
	end
	return nil;
end

-- ----------
-- Edit frame main buttons.
-- ----------

function EkBomb_EditFrame_CopyButton_OnClick()
	-- ----------
	-- Copy currently displayed warning settings into EkBomb clipboard variable.
	-- ----------
	if (not EkBomb_EditFrame_Clipboard) then
		EkBomb_EditFrame_Clipboard = {};
	end
	EkBomb_EditFrame_GetSettings(EkBomb_EditFrame_Clipboard);
	EkBomb_EditFrame_ShowMenu();
end

function EkBomb_EditFrame_CopyPaste_OnEnter(self, tip1, tip2)
	-- ----------
	-- User is over the copy or paste button.
	--
	-- Param: tip1 -- Tooltip1 text.
	-- Param: tip2 -- Tooltip2 text.
	-- ----------
	if (EkBomb_EditFrame_Clipboard) then
		tip2 = tip2 .. string.format(EkBomb_TEXT_EditFrame_CopyPaste_Current, EkBomb_EditFrame_Clipboard["name"]);
	end
	EkBomb_SetTooltip_Corner(self, tip1, tip2);
end

function EkBomb_EditFrame_PasteButton_OnClick()
	-- ----------
	-- Paste settings from EkBomb clipboard variable into currently displayed warning.
	-- ----------
	if (EkBomb_EditFrame_Clipboard) then
		EkBomb_EditFrame_PutSettings(EkBomb_EditFrame_Clipboard);
	end
end

function EkBomb_EditFrame_Ok_OnEnter(self)
	-- ----------
	-- User is over the Ok button.
	-- ----------
	if (EkBomb_EditFrame_EditMode == 1) then
		EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EditFrame_Ok_Tooltip1, EkBomb_TEXT_EditFrame_Ok_Tooltip2Edit);
	else
		EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EditFrame_Ok_Tooltip1, EkBomb_TEXT_EditFrame_Ok_Tooltip2Add);
	end
end

function EkBomb_EditFrame_OkButton_OnClick()
	-- ----------
	-- User clicked on the Ok button.
	-- ----------
	local list = EkBomb_ListFrame_Table;
	local sel  = EkBomb_ListFrame_Selected;
	local item;

	if (EkBomb_EditFrame_EditMode == 1) then
		-- Update selected item already in the list.
		item = list[sel];

	elseif (EkBomb_EditFrame_EditMode == 2) then
		-- Adding.
		-- Create new item and add to the list.
		item = {};
		tinsert(list, item);
		-- Set selection to the item being added.
		sel = #list;
	else
		-- Viewing.
		item = nil;
	end

	if (item) then
		-- Save contents of the edit fields.
		EkBomb_EditFrame_GetSettings(item);

		-- When adding, automatically enable the added item.
		if (EkBomb_EditFrame_EditMode == 2) then
			item["enabled"] = 1;
		end

		-- Sort the list
		EkBomb_ListFrame_Selected = EkBomb_ListFrame_Sort(EkBomb_ListFrame_Table, sel);

		-- Rebuild the event list entries for this table.
		EkBomb_EventList_UpdateWarningList(EkBomb_ListFrame_TableNum);
	end

	EkBomb_EditFrame:Hide();
	EkBomb_ListFrame:Show();

	-- If editing and selection number changed, or if adding...
	if ((EkBomb_EditFrame_EditMode == 1 and sel ~= EkBomb_ListFrame_Selected) or (EkBomb_EditFrame_EditMode == 2)) then
		EkBomb_ListFrame_ScrollBarUpdate();
		EkBomb_ListFrame_ScrollBarSetValue(EkBomb_ListFrame_Selected);
	end
end

function EkBomb_EditFrame_CancelButton_OnClick()
	-- ----------
	-- User clicked on the Cancel button.
	-- ----------
	EkBomb_EditFrame:Hide();
	EkBomb_ListFrame:Show();
end

-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb_EventList manipulation
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_EventList_Add(eventName, listNum, sub)
	-- ----------
	-- Add an event to the event list.
	--
	-- Param: eventName -- Name of the event (required).
	-- Param: listNum   -- List number (required).
	-- Param: sub       -- Subscript number (required).
	-- ----------
	local eventItem, eventList;
	local subList;
	local found;

	eventList = EkBomb_EventList;

	-- Using event name, get an event item.
	eventItem = eventList[eventName];
	if (not eventItem) then
		eventItem = {};
		eventItem["count"] = 0;
		eventList[eventName] = eventItem;
	end

	-- Using list number, get a subscript list.
	subList = eventItem[listNum]
	if (not subList) then
		subList = {};
		eventItem[listNum] = subList;
	end

	-- Test if subscript is already in the subscript list.
	found = nil;
	for i = 1, #subList do
		if (subList[i] == sub) then
			-- Already in the list
			found = 1;
			break;
		end
	end
	if (not found) then
		-- Add subscript to the subscript list.
		tinsert(subList, sub);

		-- Register the event and increase the count.
		if (EkBomb_Status == 1) then
			EkBomb_MessageFrame:RegisterEvent(eventName);
		end
		if (eventItem["count"] <= 0) then
			eventItem["count"] = 1;
		else
			eventItem["count"] = eventItem["count"] + 1;
		end
	end
end

function EkBomb_EventList_Delete(eventName, listNum, sub)
	-- ----------
	-- Delete an event from the event list.
	--
	-- Param: eventName -- Name of the event (required).
	-- Param: listNum   -- List number (optional).
	--                     If nil then the event will be deleted.
	-- Param: sub       -- Subscript number (optional).
	--                     If nil then the listNum will be deleted.
	-- ----------
	local eventItem, eventList;
	local subList;
	local found;

	eventList = EkBomb_EventList;

	-- Using event name, get an event item.
	eventItem = eventList[eventName];
	if (not eventItem) then
		-- Event not in list.
		return;
	end

	if (not listNum) then
		-- Delete the event.
		for i = 1, #EkBomb_WarningLists do
			eventItem[i] = nil;
		end
		eventItem["count"] = 0;
	else
		-- Using list number, get a subscript list.
		subList = eventItem[listNum];
		if (not subList) then
			-- List number does not exist.
			return;
		end

		if (not sub) then
			-- Delete the subscript list.
			eventItem["count"] = eventItem["count"] - #(eventItem[listNum]);
			eventItem[listNum] = nil;
		else
			-- Using subscript number, delete the subscript from the list.
			found = nil;
			for i = #subList, 1, -1 do
				if (subList[i] == sub) then
					tremove(subList, i);
					eventItem["count"] = eventItem["count"] - 1;
				end
			end
			if (#subList == 0) then
				eventItem[listNum] = nil;
			end
		end
	end
	
	if (eventItem["count"] <= 0) then
		eventList[eventName] = nil;
		-- Unregister the event (unless 
		EkBomb_MessageFrame:UnregisterEvent(eventName);
	end
end

function EkBomb_EventList_Update()
	-- ----------
	-- Update the event list.
	-- ----------
	for i = 1, #EkBomb_WarningLists do
		EkBomb_EventList_DeleteWarningList(i);
		EkBomb_EventList_AddWarningList(i);
	end
end

function EkBomb_EventList_DeleteAll()
	-- ----------
	-- Delete all items in the event list.
	-- ----------
	for i = 1, #EkBomb_WarningLists do
		EkBomb_EventList_DeleteWarningList(i);
	end
end

function EkBomb_EventList_UpdateWarningList(listNum)
	-- ----------
	-- Update specified warning list in the event list.
	-- ----------
	EkBomb_EventList_DeleteWarningList(listNum);
	EkBomb_EventList_AddWarningList(listNum);
end

function EkBomb_EventList_AddWarningList(listNum)
	-- ----------
	-- Add specified warning list to the appropriate events in the event list.
	-- ----------
	local list, rec;

	list = EkBomb_WarningLists[listNum];

	for i = 1, #list do
		rec = list[i];
		if (rec["enabled"] == 1) then
			EkBomb_EventList_Add(rec["event"], listNum, i);
		end
	end
end

function EkBomb_EventList_DeleteWarningList(listNum)
	-- ----------
	-- Delete specified warning list from each event (if applicable) in the event list.
	--
	-- Param: listNum -- List number (required).
	-- ----------
	local eventList;

	eventList = EkBomb_EventList;

	for eventName, eventItem in pairs(eventList) do
		if (eventItem[listNum]) then
			EkBomb_EventList_Delete(eventName, listNum);
		end
	end
end


-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------
-- *!* EkBomb_EventFrame
-- Inspired by the chat frame event link in Iriel's DevTools addon.
-- ----------------------------------------------------------------------------------------
-- ----------------------------------------------------------------------------------------

function EkBomb_EventFrame_Create()
	-- ----------
	-- Create EventFrame components.
	-- ----------
	local oframe, fs, cb, b, cbText;
	local oname = "EkBomb_EventFrame";
	local iname = "EkBomb_EventFrame_InsideFrame";
	local iframe;

	oframe = CreateFrame("Frame", oname, UIParent, "EkBomb_MainFrame_DialogBoxTemplate");
	oframe:SetFrameStrata("DIALOG");
	oframe:SetToplevel(1);
	oframe:SetClampedToScreen(1);
	oframe:SetMovable(1);
	oframe:EnableMouse();
	oframe:SetHeight(300);
	oframe:SetWidth(530);

	EkBomb_EventFrame_RestorePosition();

	oframe:SetScript("OnMouseDown", EkBomb_EventFrame_OnMouseDown);
	oframe:SetScript("OnMouseUp", EkBomb_EventFrame_OnMouseUp);
	oframe:SetScript("OnEnter", EkBomb_EventFrame_OnEnter);
	oframe:SetScript("OnLeave", EkBomb_EventFrame_OnLeave);
	oframe:SetScript("OnHide", EkBomb_EventFrame_OnHide);

	oframe:SetBackdropBorderColor(0.65, 0.65, 0.65);
	oframe:SetBackdropColor(0.15, 0.15, 0.15);

	-- Background for the various items
	iframe = CreateFrame("Frame", iname, _G[oname], "EkBomb_MainFrame_OptionFrameBoxTemplate");
	iframe:SetHeight(300);
	iframe:SetWidth(500);
	iframe:SetPoint("TOP", oname, "TOP", 0, -34);
	iframe:SetBackdropBorderColor(0.4, 0.4, 0.4);
	iframe:SetBackdropColor(0.15, 0.15, 0.15);

	-- Create up button
	b = CreateFrame("Button", "EkBomb_EventFrame_UpButton", oframe);
	b:SetHeight(24);
	b:SetWidth(24);
	b:SetPoint("TOPLEFT", oname, "TOPLEFT", 10, -10);
	b:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_UpButton_Tooltip1, EkBomb_TEXT_EventFrame_UpButton_Tooltip2, "ANCHOR_TOPLEFT"); end);
	b:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);
	b:SetScript("OnClick", EkBomb_EventFrame_UpButton_OnClick);
	b:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Up");
	b:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Down");
	b:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollUp-Disabled");
	b:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight");
	(b:GetHighlightTexture()):SetBlendMode("ADD");

	-- Create down button
	b = CreateFrame("Button", "EkBomb_EventFrame_DownButton", oframe);
	b:SetHeight(24);
	b:SetWidth(24);
	b:SetPoint("LEFT", EkBomb_EventFrame_UpButton, "RIGHT", -1, 0);
	b:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_DownButton_Tooltip1, EkBomb_TEXT_EventFrame_DownButton_Tooltip2, "ANCHOR_TOPLEFT"); end);
	b:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);
	b:SetScript("OnClick", EkBomb_EventFrame_DownButton_OnClick);
	b:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up");
	b:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down");
	b:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Disabled");
	b:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight");
	(b:GetHighlightTexture()):SetBlendMode("ADD");

	-- Title text
	oframe:CreateFontString("EkBomb_EventFrame_Title", "ARTWORK", "GameFontNormal");
	fs = EkBomb_EventFrame_Title;
	fs:SetHeight(10);
	fs:SetJustifyH("CENTER");
	fs:SetPoint("TOP", oname, "TOP", 0, -15);

	-- Close button (X).
	b = CreateFrame("Button", "EkBomb_EventFrame_CloseButton", oframe, "UIPanelCloseButton");
	b:SetPoint("TOPRIGHT", oname, "TOPRIGHT", -7, -7);

	-- Message text (when event is not available)
	iframe:CreateFontString("EkBomb_EventFrame_MessageText", "ARTWORK", "GameFontHighlight");
	fs = EkBomb_EventFrame_MessageText;
	fs:SetHeight(100);
	fs:SetWidth(iframe:GetWidth()-30);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("LEFT", iname, "LEFT", 10, 0);
	fs:SetText("");
	fs:Hide();

	-- Event name check box
	cb = CreateFrame("CheckButton", "EkBomb_EventFrame_CBEvent", iframe, "EkBomb_EventFrame_CBTemplate");
	cb:SetPoint("TOPLEFT", iname, "TOPLEFT", 10, -10);

	-- Check boxes
	for i = 1, 9 do
		cb = CreateFrame("CheckButton", "EkBomb_EventFrame_CB" .. i, iframe, "EkBomb_EventFrame_CBTemplate");
		if (i == 1) then
			cb:SetPoint("TOPLEFT", "EkBomb_EventFrame_CBEvent", "BOTTOMLEFT", 0, 0);
		else
			cb:SetPoint("TOPLEFT", "EkBomb_EventFrame_CB" .. i-1, "BOTTOMLEFT", 0, 0);
		end
		cb:SetID(i);
		cbText = _G[cb:GetName() .. "_Text"];
		cbText:SetHeight(10);
		cbText:SetWidth(iframe:GetWidth() - 50);
		cbText:SetJustifyH("LEFT");
	end

	-- "Search type"
	iframe:CreateFontString("EkBomb_EventFrame_SearchTypeText", "ARTWORK", "GameFontNormal");
	fs = EkBomb_EventFrame_SearchTypeText;
	fs:SetHeight(10);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("TOPLEFT", iname, "BOTTOMLEFT", 13, 60);
	fs:SetText(EkBomb_TEXT_EventFrame_SearchTypeText);

	cb = CreateFrame("CheckButton", "EkBomb_EventFrame_ExactCB", iframe, "OptionsCheckButtonTemplate");
	cb:SetPoint("LEFT", "EkBomb_EventFrame_SearchTypeText", "RIGHT", 10, 0);
	cb:SetHeight(22);
	cb:SetWidth(22);
	cb:SetID(101);
	cb:SetHitRectInsets(0, 0, 0, 0);
	cb:SetScript("OnClick", EkBomb_EventFrame_ExactCB_OnClick);
	cb:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_ExactCB_Tooltip1, EkBomb_TEXT_EventFrame_ExactCB_Tooltip2); end);
	cb:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);

	iframe:CreateFontString("EkBomb_EventFrame_ExactText", "ARTWORK", "GameFontHighlight");
	fs = EkBomb_EventFrame_ExactText;
	fs:SetHeight(10);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("LEFT", "EkBomb_EventFrame_ExactCB", "RIGHT", 0, 0);
	fs:SetText(EkBomb_TEXT_EventFrame_ExactText);

	cb = CreateFrame("CheckButton", "EkBomb_EventFrame_PatternCB", iframe, "OptionsCheckButtonTemplate");
	cb:SetPoint("LEFT", "EkBomb_EventFrame_ExactText", "RIGHT", 10, 0);
	cb:SetHeight(22);
	cb:SetWidth(22);
	cb:SetID(102);
	cb:SetHitRectInsets(0, 0, 0, 0);
	cb:SetScript("OnClick", EkBomb_EventFrame_PatternCB_OnClick);
	cb:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_PatternCB_Tooltip1, EkBomb_TEXT_EventFrame_PatternCB_Tooltip2); end);
	cb:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);

	iframe:CreateFontString("EkBomb_EventFrame_PatternText", "ARTWORK", "GameFontHighlight");
	fs = EkBomb_EventFrame_PatternText;
	fs:SetHeight(10);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("LEFT", "EkBomb_EventFrame_PatternCB", "RIGHT", 0, 0);
	fs:SetText(EkBomb_TEXT_EventFrame_PatternText);

	EkBomb_EventFrame_SearchTypeText:Hide();
	EkBomb_EventFrame_ExactCB:Hide();
	EkBomb_EventFrame_ExactText:Hide();
	EkBomb_EventFrame_PatternCB:Hide();
	EkBomb_EventFrame_PatternText:Hide();

	-- "Create warning:"
	iframe:CreateFontString("EkBomb_EventFrame_CreateText", "ARTWORK", "GameFontNormal");
	fs = EkBomb_EventFrame_CreateText;
	fs:SetHeight(10);
	fs:SetJustifyH("LEFT");
	fs:SetPoint("TOPLEFT", "EkBomb_EventFrame_SearchTypeText", "BOTTOMLEFT", 0, -20);
	fs:SetText(EkBomb_TEXT_EventFrame_CreateText);

	-- Create General button
	b = CreateFrame("Button", "EkBomb_EventFrame_GeneralButton", iframe, "UIPanelButtonTemplate");
	b:SetHeight(30);
	b:SetWidth(140);
	b:SetPoint("LEFT", "EkBomb_EventFrame_CreateText", "RIGHT", 10, 0);
	b:SetText(EkBomb_TEXT_EventFrame_GeneralButton_Text);
	b:Disable();
	b:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_GeneralButton_Tooltip1, EkBomb_TEXT_EventFrame_GeneralButton_Tooltip2); end);
	b:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);
	b:SetScript("OnClick", EkBomb_EventFrame_GeneralButton_OnClick);

	-- Create Character button
	b = CreateFrame("Button", "EkBomb_EventFrame_CharacterButton", iframe, "UIPanelButtonTemplate");
	b:SetHeight(30);
	b:SetWidth(140);
	b:SetPoint("LEFT", "EkBomb_EventFrame_GeneralButton", "RIGHT", 10, 0);
	b:SetText(EkBomb_TEXT_EventFrame_CharacterButton_Text);
	b:Disable();
	b:SetScript("OnEnter", function(self) EkBomb_SetTooltip_Corner(self, EkBomb_TEXT_EventFrame_CharacterButton_Tooltip1, EkBomb_TEXT_EventFrame_CharacterButton_Tooltip2); end);
	b:SetScript("OnLeave", function() EkBomb_HideTooltip(); end);
	b:SetScript("OnClick", EkBomb_EventFrame_CharacterButton_OnClick);

end

function EkBomb_EventFrame_Reset()
	-- ----------
	-- Reset event frame position.
	-- ----------	
	EkBomb_EventFrame:ClearAllPoints();
	EkBomb_EventFrame:SetPoint("CENTER", "UIParent", "CENTER", 0, 0);
	EkBomb_EventFrame_SavePosition();
end

function EkBomb_EventFrame_SavePosition()
	-- ----------------
	-- Save current position of the event frame.
	-- ----------------
	local ap, rf, rp, rx, ry = EkBomb_EventFrame:GetPoint();
	if (rf) then
		rf = rf:GetName();
	end
	if (type(EkBomb_EventPosition) ~= "table") then
		EkBomb_EventPosition = {};
	end
	EkBomb_EventPosition.ap = ap;
	EkBomb_EventPosition.rf = rf;
	EkBomb_EventPosition.rp = rp;
	EkBomb_EventPosition.rx = rx;
	EkBomb_EventPosition.ry = ry;
end

function EkBomb_EventFrame_RestorePosition()
	-- ----------------
	-- Restore current position of the event frame.
	-- ----------------
	local pos = EkBomb_EventPosition;
	if (pos) then
		EkBomb_EventFrame:ClearAllPoints();
		EkBomb_EventFrame:SetPoint(pos.ap, pos.rf, pos.rp, pos.rx, pos.ry);
	else
		EkBomb_EventFrame_Reset();
	end
end

function EkBomb_EventFrame_HookFunctions()
	-- ----------
	-- Hook some functions.
	-- ----------
	local frame;

	-- Hook chat frames' AddMessage function so we can add our own link to each chat line.
	for f = 1, NUM_CHAT_WINDOWS do
		frame = _G["ChatFrame" .. f];
		if (frame and not EkBomb_ChatFrame_Old[frame]) then
			EkBomb_ChatFrame_Old[frame] = frame.AddMessage;
			frame.AddMessage = EkBomb_ChatFrame_AddMessage;
		end
	end

	-- Hook the SetItemRef() function so that when the user clicks on a link we can
	-- test for the links we added to the chat line.
	if (not EkBomb_SetItemRef_Old) then
		EkBomb_SetItemRef_Old = SetItemRef;
		SetItemRef = EkBomb_SetItemRef;
	end
end

function EkBomb_ChatFrame_AddMessage(self, msg, r, g, b, id)
	-- ----------
	-- Hooked chat frame AddMessage function.
	-- ----------
	local eventNum, item, index, first, arg;
	local linkText;

	if (EkBomb_ShowLink == 1) then
		linkText = "E";

		-- NOTE: In WoW 4.0 the global event variable will be removed,
		-- so this will no longer work.
		if (event) then
			-- Next event number to use
			eventNum = EkBomb_EventFrame_LastEventNum + 1;

			-- Update last event number
			EkBomb_EventFrame_LastEventNum = eventNum;

			-- Update first event number
			first = EkBomb_EventFrame_LastEventNum - EkBomb_EventBufferSize + 1;
			if (first < 1) then
				first = 1;
			end
			if (first > EkBomb_EventFrame_FirstEventNum) then
				EkBomb_EventFrame_FirstEventNum = first;
			end

			EkBomb_EventFrame_UpdateEventNum();

			-- Calculate buffer index (1 to buffersize)
			index = math.fmod(eventNum - 1, EkBomb_EventBufferSize) + 1;

			-- Save the event item
			item = EkBomb_EventFrame_Buffer[index];
			if (not item) then
				item = {};
			end
			item[1] = event;  -- Won't work in WoW 4.0 since the global event variable will be removed.
			for i = 2, 10 do
				-- NOTE: In WoW 4.0 the global arg1, arg2, etc. variables will be removed,
				-- so this will no longer work.
				arg = _G["arg" .. i-1];
				-- Only interested in strings with a value.
				if (arg and type(arg) == "string" and arg ~= "") then
					item[i] = arg;
				end
			end

			EkBomb_EventFrame_Buffer[index] = item;

			if (EkBomb_TestFlag == 1) then
				linkText = eventNum;
			end

			-- Add a link to the chat text.
			msg = "|cffffdd00|HEkBomb:en:" .. eventNum .. "|h[" .. linkText .. "]|h|r " .. msg;
		else
			-- Add a non-link to the chat text.
			msg = "|cff888888[" .. linkText .. "]|r " .. msg;
		end
	end
	return EkBomb_ChatFrame_Old[self](self, msg, r, g, b, id);
end

function EkBomb_SetItemRef(link, text, button)
	-- ----------
	-- Hooked SetItemRef function.
	-- ----------
	local pos1, pos2;
	local eventNum, first;
	local item, index;

	if (strsub(link, 1, 7) ~= "EkBomb:") then
		-- Not our link, so call the old function.
		return EkBomb_SetItemRef_Old(link, text, button);
	end

	-- Extract the event number
	pos1, pos2, eventNum = string.find(link, "^EkBomb:en:(%d+)$");
	if (not eventNum) then
		return;
	end

	-- Create the EventFrame if needed.
	if (not EkBomb_EventFrame) then
		EkBomb_EventFrame_Create();
	end

	-- Calculate first possible event number in the buffer.
	first = EkBomb_EventFrame_LastEventNum - EkBomb_EventBufferSize + 1;
	if (first < 1) then
		first = 1;
	end

	-- Check if event number is still in the buffer
	eventNum = tonumber(eventNum);
	if (eventNum < first) then
		-- Event is no longer in the buffer.
		EkBomb_EventFrame_ShowMessage(eventNum, EkBomb_TEXT_EventFrame_OldEvent);
		return;
	end

	-- Get event item from buffer
	index = math.fmod(eventNum - 1, EkBomb_EventBufferSize) + 1;
	item = EkBomb_EventFrame_Buffer[index];

	-- Event may (or may not) be in the buffer.
	if (not item) then
		-- No event data available for this event number
		-- The buffer size may have been recently increased by the user, resulting
		-- in some nil entries in the table for old event numbers which scrolled out
		-- of the old buffer but now fit in the new buffer.
		EkBomb_EventFrame_ShowMessage(eventNum, EkBomb_TEXT_EventFrame_NoEvent);
		return;
	end

	EkBomb_EventFrame_ShowEvent(eventNum, item);
end

function EkBomb_EventFrame_ResizeBuffer(newSize)
	-- ----------
	-- Resize the event buffer.
	--
	-- Param: newSize -- New size of the buffer (minimum 1).
	-- ----------
	local oldIndex, newIndex, item;
	local newBuffer;

	-- If new size is same as the old size, or is zero...
	if (newSize == EkBomb_EventBufferSize or newSize == 0) then
		return;
	end

	-- If no events yet...
	if (EkBomb_EventFrame_LastEventNum == 0) then
		EkBomb_EventBufferSize = newSize;
		return;
	end

	-- If the new buffer is larger than the old one...
	if (newSize > EkBomb_EventBufferSize) then

		-- If we have never filled the older buffer..
		if (EkBomb_EventFrame_LastEventNum <= EkBomb_EventBufferSize) then
			EkBomb_EventBufferSize = newSize;
			return;
		end
	else
		-- The new buffer is smaller than the old...
		-- Calculate the first event number in new buffer.
		EkBomb_EventFrame_FirstEventNum = EkBomb_EventFrame_LastEventNum - newSize + 1;
		if (EkBomb_EventFrame_FirstEventNum < 1) then
			EkBomb_EventFrame_FirstEventNum = 1;
		end
	end

	-- Transfer the items from the old buffer to the new buffer.
	newBuffer = {};

	for i = EkBomb_EventFrame_FirstEventNum, EkBomb_EventFrame_LastEventNum do
		-- Old index
		oldIndex = math.fmod(i - 1, EkBomb_EventBufferSize) + 1;

		-- New index
		newIndex = math.fmod(i - 1, newSize) + 1;

		-- Transfer the item
		item = EkBomb_EventFrame_Buffer[oldIndex];
		if (item) then
			newBuffer[newIndex] = item;
		end
	end

	-- Use the new buffer and size.
	EkBomb_EventFrame_Buffer = newBuffer;
	EkBomb_EventBufferSize = newSize;

	EkBomb_EventFrame_UpdateEventNum();
end

function EkBomb_EventFrame_UpButton_OnClick()
	-- ----------
	-- User clicked on the 'up' button.
	-- Show previous (older) event from the buffer.
	-- ----------
	local eventNum, index, item;

	-- The currently displayed event may be:
	--   a) An event that is no longer in the buffer
	--   b) An event that is in the buffer

	-- Get currently displayed event number
	eventNum = EkBomb_EventFrame.ekEventNum;

	-- If the currently displayed event is no longer in the buffer...
	if (eventNum < EkBomb_EventFrame_FirstEventNum) then
		return;
	end

	-- Advance to previous (older) event.
	eventNum = eventNum - 1;

	-- If the currently displayed event is no longer in the buffer...
	if (eventNum < EkBomb_EventFrame_FirstEventNum) then
		return;
	end

	if (IsAltKeyDown()) then
		eventNum = EkBomb_EventFrame_FirstEventNum;
	elseif (IsShiftKeyDown()) then
		eventNum = eventNum - 9;
	elseif (IsControlKeyDown()) then
		eventNum = eventNum - 99;
	end
	if (eventNum < EkBomb_EventFrame_FirstEventNum) then
		eventNum = EkBomb_EventFrame_FirstEventNum;
	end

	index = math.fmod(eventNum - 1, EkBomb_EventBufferSize) + 1;
	item = EkBomb_EventFrame_Buffer[index];
	if (not item) then
		return;
	end

	-- Show the event
	EkBomb_EventFrame_ShowEvent(eventNum, item);
end

function EkBomb_EventFrame_DownButton_OnClick()
	-- ----------
	-- User clicked on the 'down' button.
	-- Show next (more recent) event from the buffer.
	-- ----------
	local eventNum, index, item;

	-- The currently displayed event may be:
	--   a) An event that is no longer in the buffer
	--   b) An event that is in the buffer

	-- Get currently displayed event number.
	eventNum = EkBomb_EventFrame.ekEventNum;

	if (eventNum < EkBomb_EventFrame_FirstEventNum) then
		-- Use the first event in the buffer.
		eventNum = EkBomb_EventFrame_FirstEventNum;
	else
		-- Advance to the next (more recent) event.
		eventNum = eventNum + 1;

		-- If we were already at the last event...
		if (eventNum > EkBomb_EventFrame_LastEventNum) then
			return;
		end
	end

	if (IsAltKeyDown()) then
		eventNum = EkBomb_EventFrame_LastEventNum;
	elseif (IsShiftKeyDown()) then
		eventNum = eventNum + 9;
	elseif (IsControlKeyDown()) then
		eventNum = eventNum + 99;
	end
	if (eventNum > EkBomb_EventFrame_LastEventNum) then
		eventNum = EkBomb_EventFrame_LastEventNum;
	end

	index = math.fmod(eventNum - 1, EkBomb_EventBufferSize) + 1;
	item = EkBomb_EventFrame_Buffer[index];
	if (not item) then
		return;
	end

	-- Show the event
	EkBomb_EventFrame_ShowEvent(eventNum, item);
end

function EkBomb_EventFrame_UpdateEventNum()
	-- ----------
	-- Update the event number.
	--
	-- Param: eventNum -- Event number
	-- ----------
	if (not EkBomb_EventFrame) then
		return;
	end
	if (EkBomb_EventFrame:IsShown()) then
		local eventNum = EkBomb_EventFrame.ekEventNum;
		local size = EkBomb_EventFrame_LastEventNum - EkBomb_EventFrame_FirstEventNum + 1;
		local curr = size - (EkBomb_EventFrame_LastEventNum - eventNum);
		if (curr < 1) then
			curr = 0;
		end
		-- Event 62  (2 / 10)
		EkBomb_EventFrame_Title:SetText(string.format("%s %d  (%d / %d)", EkBomb_TEXT_EventFrame_EventTitle, eventNum, curr, size));
	end
end

function EkBomb_EventFrame_ShowMessage(eventNum, msg)
	-- ----------
	-- Display a message in the event window.
	--
	-- Param: eventNum -- Event number
	-- Param: msg -- Message text
	-- ----------
	local cb;

	EkBomb_EventFrame.ekEventNum = eventNum;

	EkBomb_ReAnchor_TopLeft_TopLeft(EkBomb_EventFrame);

	EkBomb_EventFrame_MessageText:SetText(msg);
	EkBomb_EventFrame_MessageText:Show();

	EkBomb_EventFrame_CBEvent:Hide();
	for i = 1, 9 do
		cb = _G["EkBomb_EventFrame_CB" .. i];
		cb:Hide();
	end

	EkBomb_EventFrame_SearchTypeText:Hide();
	EkBomb_EventFrame_ExactCB:Hide();
	EkBomb_EventFrame_ExactText:Hide();
	EkBomb_EventFrame_PatternCB:Hide();
	EkBomb_EventFrame_PatternText:Hide();
	EkBomb_EventFrame_CreateText:Hide();
	EkBomb_EventFrame_GeneralButton:Hide();
	EkBomb_EventFrame_CharacterButton:Hide();

	EkBomb_EventFrame:SetHeight(170);
	EkBomb_EventFrame_InsideFrame:SetHeight(EkBomb_EventFrame:GetHeight() - 50);

	EkBomb_EventFrame:Show();
	EkBomb_EventFrame:Raise();

	EkBomb_EventFrame_UpdateEventNum();
end

function EkBomb_EventFrame_ShowEvent(eventNum, item)
	-- ----------
	-- Display an event in the event window.
	--
	-- Param: eventNum -- Event number
	-- Param: item -- Event item array
	-- ----------
	local count;
	local index;
	local cb, cbText, arg;
	local cbHeight, fSize;

	EkBomb_EventFrame.ekEventNum = eventNum;

	EkBomb_ReAnchor_TopLeft_TopLeft(EkBomb_EventFrame);

	EkBomb_EventFrame_MessageText:Hide();

	-- Event name
	EkBomb_EventFrame_CBEvent_Text:SetText("|cffffdd00" .. EkBomb_TEXT_EventFrame_Event .. ":|r  " .. item[1]);
	EkBomb_EventFrame_CBEvent.ekTip1 = EkBomb_TEXT_EventFrame_Event_Tooltip1;
	EkBomb_EventFrame_CBEvent.ekTip2 = EkBomb_TEXT_EventFrame_Event_Tooltip2;
	EkBomb_EventFrame_CBEvent.ekEvent = item[1];
	EkBomb_EventFrame_CBEvent:SetChecked(0);
	EkBomb_EventFrame_CBEvent:Show();

	-- Event arguments
	count = 0;
	for i = 2, 10 do
		if (item[i]) then
			count = count + 1;
			cbText = _G["EkBomb_EventFrame_CB" .. count .. "_Text"];

			-- Convert certain characters so they display properly
			arg = string.gsub(item[i], "|",  "||");
			arg = string.gsub(arg,     "\n", "\\n");
			-- arg = string.gsub(arg,     "\"", "\\\"");

			cbText:SetText("|cffffdd00" .. EkBomb_TEXT_EventFrame_Arg .. " " .. i-1 .. ":|r  " .. arg);

			cb = _G["EkBomb_EventFrame_CB" .. count];
			cb.ekTip1 = EkBomb_TEXT_EventFrame_Arg_Tooltip1;
			cb.ekTip2 = EkBomb_TEXT_EventFrame_Arg_Tooltip2;
			cb.ekArg = item[i];  -- raw event value
			cb:SetChecked(0);
			cb:Show();
		end
	end

	cbHeight = EkBomb_EventFrame_CB1:GetHeight();
	fSize = cbHeight + (count * cbHeight) + 130;
	EkBomb_EventFrame:SetHeight(fSize);
	EkBomb_EventFrame_InsideFrame:SetHeight(fSize - 50);

	while (count < 9) do
		count = count + 1;
		cb = _G["EkBomb_EventFrame_CB" .. count];
		cb:SetChecked(0);
		cb:Hide();
	end

	EkBomb_EventFrame_SearchTypeText:Hide();
	EkBomb_EventFrame_ExactCB:Hide();
	EkBomb_EventFrame_ExactText:Hide();
	EkBomb_EventFrame_PatternCB:Hide();
	EkBomb_EventFrame_PatternText:Hide();

	EkBomb_EventFrame_CreateText:Show();
	EkBomb_EventFrame_GeneralButton:Show();
	EkBomb_EventFrame_CharacterButton:Show();

	EkBomb_EventFrame:Show();
	EkBomb_EventFrame:Raise();

	EkBomb_EventFrame_UpdateEventNum();
end

function EkBomb_EventFrame_OnHide()
	-- ----------
	-- Event frame is being hidden.
	-- ----------
	EkBomb_EventFrame:StopMovingOrSizing();
	EkBomb_EventFrame_SavePosition();
end

function EkBomb_EventFrame_OnMouseDown(self, button)
	-- ----------
	-- User pressed mouse button on the event frame.
	-- ----------
	if ( button == "LeftButton" ) then
		EkBomb_EventFrame:StartMoving();
	elseif ( button == "RightButton" ) then
		EkBomb_EventFrame_Reset();
	end
end

function EkBomb_EventFrame_OnMouseUp(self, button)
	-- ----------
	-- User released mouse button on the event frame.
	-- ----------
	if ( button == "LeftButton" ) then
		EkBomb_EventFrame:StopMovingOrSizing();
		EkBomb_EventFrame_SavePosition();
	end
end

function EkBomb_EventFrame_OnEnter(self)
	-- ----------
	-- Mouse is over the event frame.
	-- ----------
	EkBomb_SetTooltip_Anchor(self, EkBomb_TEXT_Warning_Drag_Tooltip, "ANCHOR_RIGHT");
end

function EkBomb_EventFrame_OnLeave()
	-- ----------
	-- Mouse has left the event frame.
	-- ----------
	EkBomb_HideTooltip();
end

function EkBomb_EventFrame_CB_OnClick(self)
	-- ----------
	-- User clicked on an event frame check box.
	-- ----------
	local cb;
	local count;

	count = 0;

	EkBomb_EventFrame_SearchTypeText:Hide();
	EkBomb_EventFrame_ExactCB:Hide();
	EkBomb_EventFrame_ExactText:Hide();
	EkBomb_EventFrame_PatternCB:Hide();
	EkBomb_EventFrame_PatternText:Hide();

	if (self:GetChecked()) then
		-- Uncheck all the other checkboxes.
		if (self:GetName() ~= "EkBomb_EventFrame_CBEvent") then
			EkBomb_EventFrame_CBEvent:SetChecked(0);
		end
		for i = 1, 9 do
			cb = _G["EkBomb_EventFrame_CB" .. i];
			if (cb ~= self) then
				cb:SetChecked(0);
			end
		end

		-- We just checked a checkbox, so just that one is now checked.
		EkBomb_EventFrame_GeneralButton:Enable();
		EkBomb_EventFrame_CharacterButton:Enable();

		if (self:GetName() ~= "EkBomb_EventFrame_CBEvent") then
			EkBomb_EventFrame_SearchTypeText:Show();
			EkBomb_EventFrame_ExactCB:Show();
			EkBomb_EventFrame_ExactText:Show();
			EkBomb_EventFrame_PatternCB:Show();
			EkBomb_EventFrame_PatternText:Show();

			-- If the player's name is in the argument text, then default to search pattern.
			if (string.find(self.ekArg, (UnitName("player")))) then
				EkBomb_EventFrame_ExactCB:SetChecked(0);
				EkBomb_EventFrame_PatternCB:SetChecked(1);
			else
				-- Player's name is not in the argument, so default to an exact search.
				EkBomb_EventFrame_ExactCB:SetChecked(1);
				EkBomb_EventFrame_PatternCB:SetChecked(0);
			end
		end
	else
		-- We just unchecked a checkbox, so nothing is left checked.
		EkBomb_EventFrame_GeneralButton:Disable();
		EkBomb_EventFrame_CharacterButton:Disable();
	end
end

function EkBomb_EventFrame_ExactCB_OnClick()
	-- ----------
	-- User clicked on the "exact" search type checkbox.
	-- ----------
	local cb = EkBomb_EventFrame_ExactCB;
	local cb2 = EkBomb_EventFrame_PatternCB;
	local other = 1;
	if (cb:GetChecked()) then
		other = 0;
	end
	cb2:SetChecked(other);
end

function EkBomb_EventFrame_PatternCB_OnClick()
	-- ----------
	-- User clicked on the "pattern" search type checkbox.
	-- ----------
	local cb = EkBomb_EventFrame_PatternCB;
	local cb2 = EkBomb_EventFrame_ExactCB;
	local other = 1;
	if (cb:GetChecked()) then
		other = 0;
	end
	cb2:SetChecked(other);
end

function EkBomb_EventFrame_GeneralButton_OnClick()
	-- ----------
	-- User clicked on the button to create a general warning.
	-- ----------
	EkBomb_EventFrame_CreateWarning(EkBomb_ListFrame_TabGeneral);
end

function EkBomb_EventFrame_CharacterButton_OnClick()
	-- ----------
	-- User clicked on the button to create a per character warning.
	-- ----------
	EkBomb_EventFrame_CreateWarning(EkBomb_ListFrame_TabCharacter);
end

function EkBomb_EventFrame_CreateWarning(tab)
	-- ----------
	-- Create a custom warning based on user's selections in the event frame.
	--
	-- Param: tab -- Which tab number to use on the custom frame.
	-- ----------
	local argNum, argText, cb;
	local seachCode, searchText;
	local playerCode;
	local eventName;
	local item;

	-- Show the general tab of the custom list frame.
	if (not EkBomb_MainFrame:IsShown()) then
		EkBomb_MainFrame:Show();
	end
	EkBomb_MainFrame_GotoTab(EkBomb_MainFrame_TabCustom);
	EkBomb_ListFrame_Tab_OnClick(tab);

	-- Hide the list frame prior to showing the edit frame.
	EkBomb_ListFrame:Hide();

	-- Get name of the event
	eventName = EkBomb_EventFrame_CBEvent.ekEvent;

	-- Determine selected item (event or event argument)
	argNum = 0;  -- Default to event with no argument
	argText = "";
	for i = 1, 9 do
		cb = _G["EkBomb_EventFrame_CB" .. i];
		if (cb:GetChecked()) then
			argNum = i;  -- Event argument is selected
			argText = cb.ekArg;
			break;
		end
	end

	if (EkBomb_EventFrame_PatternCB:GetChecked()) then
		-- Create a search pattern.
		searchCode = EkBomb_SearchCode_Pattern;
		searchText = argText;

		-- Escape any special characters in the pattern (eg. replace "$" with "%$", etc)
		searchText = string.gsub(searchText, "%%", "%%%%");
		searchText = string.gsub(searchText, "%^", "%%%^");
		searchText = string.gsub(searchText, "%$", "%%%$");
		searchText = string.gsub(searchText, "%(", "%%%(");
		searchText = string.gsub(searchText, "%)", "%%%)");
		searchText = string.gsub(searchText, "%.", "%%%.");
		searchText = string.gsub(searchText, "%[", "%%%[");
		searchText = string.gsub(searchText, "%]", "%%%]");
		searchText = string.gsub(searchText, "%*", "%%%*");
		searchText = string.gsub(searchText, "%+", "%%%+");
		searchText = string.gsub(searchText, "%-", "%%%-");
		searchText = string.gsub(searchText, "%?", "%%%?");

		-- Anchor the match to the start and end of the argument text.
		searchText = "^" .. searchText .. "$";

		-- If player name is in the string, replace it with (%S+)
		local pos1, pos2 = string.find(searchText, (UnitName("player")));
		if (pos1 and pos2) then
			searchText = strsub(searchText, 1, pos1-1) .. "(%S+)" .. strsub(searchText, pos2+1);
		end
	else
		-- Create an exact match search.
		searchCode = EkBomb_SearchCode_Exact;
		searchText = argText;
	end

	-- Initialize the edit fields
	item = {};
	EkBomb_EditFrame_NewSettings(item);
	item["event"]  = eventName;
	item["arg"]    = argNum;
	item["search"] = searchCode;
	item["value"]  = searchText;
	EkBomb_EditFrame_PutSettings(item);

	EkBomb_EditFrame_EditMode = 2;  -- 0==View, 1==Edit, 2==Add

	EkBomb_EditFrame:Show();
	EkBomb_EditFrame:Raise();
	EkBomb_EditFrame_NameEB:SetFocus();
end
